# 7 "parsing_cocci/lexer_cocci.mll"
 
open Parser_cocci_menhir
module D = Data
module Ast = Ast_cocci
module Ast0 = Ast0_cocci
module P = Parse_aux
module FC = Flag_parsing_cocci
exception Lexical of string
let tok = Lexing.lexeme

let line = ref 1
let file = ref ""
let logical_line = ref 0

(* ---------------------------------------------------------------------- *)
(* control codes *)

(* Defined in data.ml
type line_type = MINUS | OPTMINUS | PLUS | CONTEXT | OPT
*)

let current_line_type = ref (D.CONTEXT,!line,!logical_line)

let prev_plus = ref false
let line_start = ref 0 (* offset of the beginning of the line *)
let get_current_line_type lexbuf =
  let (c,l,ll) = !current_line_type in
  let lex_start = Lexing.lexeme_start lexbuf in
  let preceeding_spaces =
    if !line_start < 0 then 0 else lex_start - !line_start in
  (*line_start := -1;*)
  prev_plus := (c = D.PLUS) || (c = D.PLUSPLUS);
  (c,l,ll,ll,lex_start,preceeding_spaces,[],[],[],"")
let current_line_started = ref false
let col_zero = ref true

let contextify (c,l,ll,lle,lex_start,preceeding_spaces,bef,aft,pos,ws) =
  (D.CONTEXT,l,ll,lle,lex_start,preceeding_spaces,bef,aft,pos,ws)

let reset_line lexbuf =
  line := !line + 1;
  current_line_type := (D.CONTEXT,!line,!logical_line);
  current_line_started := false;
  (if not(!Data.in_meta || !Data.in_rule_name) then col_zero := true);
  line_start := Lexing.lexeme_start lexbuf + 1

let started_line = ref (-1)

let start_line seen_char =
  current_line_started := true;
  col_zero := false;
  (if seen_char && not(!line = !started_line)
  then
    begin
      started_line := !line;
      logical_line := !logical_line + 1
    end)

let pass_zero _ = col_zero := false

let lexerr s1 s2 = raise (Lexical (Printf.sprintf "%s%s" s1 s2))

let opt_reverse_token token =
  if !FC.interpret_inverted
  then match token with
         D.MINUS        -> D.PLUSPLUS  (* maybe too liberal *)
       | D.OPTMINUS     -> lexerr "cannot invert token ?- (an optional minus line), which is needed for reversing the patch" ""
       | D.PLUS         -> D.MINUS
       | D.PLUSPLUS     -> D.MINUS (* may not be sufficient *)
       | _              -> token
  else token

let add_current_line_type x =
  match (opt_reverse_token x,!current_line_type) with
    (D.MINUS,(D.CONTEXT,ln,lln))  ->
      current_line_type := (D.MINUS,ln,lln)
  | (D.MINUS,(D.OPT,ln,lln))      ->
      current_line_type := (D.OPTMINUS,ln,lln)
  | (D.PLUS,(D.CONTEXT,ln,lln))   ->
      current_line_type := (D.PLUS,ln,lln)
  | (D.PLUSPLUS,(D.CONTEXT,ln,lln))   ->
      current_line_type := (D.PLUSPLUS,ln,lln)
  | (D.OPT,(D.CONTEXT,ln,lln))    ->
      current_line_type := (D.OPT,ln,lln)
  | _ -> lexerr "invalid control character combination" ""

let check_minus_context_linetype s =
  match !current_line_type with
    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) -> lexerr "invalid in a + context: " s
  | _ -> ()

let check_context_linetype s =
  match !current_line_type with
    (D.CONTEXT,_,_) -> ()
  | _ -> lexerr "invalid in a nonempty context: " s

let check_plus_linetype s =
  match !current_line_type with
    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) -> ()
  | _ -> lexerr "invalid in a non + context: " s

let check_arity_context_linetype s =
  match !current_line_type with
    (D.CONTEXT,_,_) | (D.PLUS,_,_) | (D.PLUSPLUS,_,_)
  | (D.OPT,_,_) -> ()
  | _ -> lexerr "invalid in a nonempty context: " s

let check_comment s =
  if not !current_line_started
  then lexerr "+ expected at the beginning of the line" s

let process_include start finish str =
  (match !current_line_type with
    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
      (try
	let _ = Str.search_forward (Str.regexp "\\.\\.\\.") str start in
	lexerr "... not allowed in + include" ""
      with Not_found -> ())
  | _ -> ());
  String.sub str (start + 1) (finish - start - 1)

(* ---------------------------------------------------------------------- *)
type pm = PATCH | MATCH | UNKNOWN

let pm = ref UNKNOWN

let patch_or_match = function
    PATCH ->
      if not !D.ignore_patch_or_match
      then
	(match !pm with
	  MATCH ->
	    lexerr "- or + not allowed in the first column for a match" ""
	| PATCH -> ()
	| UNKNOWN -> Flag.sgrep_mode2 := false; pm := PATCH)
  | MATCH ->
      if not !D.ignore_patch_or_match
      then
	(match !pm with
	  PATCH -> lexerr "* not allowed in the first column for a patch" ""
	| MATCH -> ()
	| UNKNOWN -> Flag.sgrep_mode2 := true; pm := MATCH)
  | _ -> failwith "unexpected argument"

(* ---------------------------------------------------------------------- *)
(* identifiers, including metavariables *)

let metavariables = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let all_metavariables =
  (Hashtbl.create(100) : (string,(string * (D.clt -> token)) list) Hashtbl.t)

let type_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let attr_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let declarer_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let iterator_names = (Hashtbl.create(100) : (string, D.clt -> token) Hashtbl.t)

let symbol_names = (Hashtbl.create(15) : (string, D.clt -> token) Hashtbl.t)

let rule_names = (Hashtbl.create(100) : (string, unit) Hashtbl.t)

let check_var s linetype =
  let fail _ =
    if (!Data.in_prolog || !Data.in_rule_name) &&
      Str.string_match (Str.regexp "<.*>") s 0
    then TPathIsoFile s
    else
      try (Hashtbl.find metavariables s) linetype
      with Not_found ->
	(try (Hashtbl.find type_names s) linetype
	with Not_found ->
	  (try (Hashtbl.find attr_names s) linetype
	  with Not_found ->
	    (try (Hashtbl.find declarer_names s) linetype
	    with Not_found ->
	      (try (Hashtbl.find iterator_names s) linetype
	      with Not_found ->
		(try (Hashtbl.find symbol_names s) linetype
		with Not_found ->
                  TIdent (s,linetype)))))) in
  if !Data.in_meta || !Data.in_rule_name
  then (try Hashtbl.find rule_names s; TRuleName s with Not_found -> fail())
  else fail()

let id_tokens lexbuf =
  let s = tok lexbuf in
  let linetype = get_current_line_type lexbuf in
  let in_rule_name = !Data.in_rule_name in
  let in_meta = !Data.in_meta && not !Data.saw_struct in
  let in_iso = !Data.in_iso in
  let in_prolog = !Data.in_prolog in
  (if s = "identifer" && in_meta
  then Common.pr2 "Warning: should identifer be identifier?");
  match s with
    "metavariable" when in_meta -> check_arity_context_linetype s; TMetavariable
  | "identifier" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TIdentifier
  | "type" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TType
  | "parameter" when in_meta ->  check_arity_context_linetype s; TParameter
  | "operator" when in_meta ->   check_arity_context_linetype s; TOperator
  | "binary" when in_meta ->   check_arity_context_linetype s; TBinary
  | "assignment" when in_meta ->   check_arity_context_linetype s; TAssignment
  | "constant"  when in_meta ->  check_arity_context_linetype s; TConstant
  | "generated" when in_rule_name && not (!Flag.make_hrule = None) ->
      check_arity_context_linetype s; TGenerated
  | "expression" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TExpression
  | "declaration" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TDeclaration
  | "field" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TField
  | "initialiser" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TInitialiser
  | "initializer" when in_meta || in_rule_name ->
      check_arity_context_linetype s; TInitialiser
  | "idexpression" when in_meta ->
      check_arity_context_linetype s; TIdExpression
  | "statement" when in_meta ->  check_arity_context_linetype s; TStatement
  | "function"  when in_meta ->  check_arity_context_linetype s; TFunction
  | "local" when in_meta ->      check_arity_context_linetype s; TLocal
  | "global" when in_meta ->     check_arity_context_linetype s; TGlobal
  | "list" when in_meta ->       check_arity_context_linetype s; Tlist
  | "fresh" when in_meta ->      check_arity_context_linetype s; TFresh
  | "typedef" when in_meta ->    check_arity_context_linetype s; TTypedef
  | "attribute" when in_meta ->  check_arity_context_linetype s; TAttribute
  | "declarer" when in_meta ->   check_arity_context_linetype s; TDeclarer
  | "iterator" when in_meta ->   check_arity_context_linetype s; TIterator
  | "name" when in_meta ->       check_arity_context_linetype s; TName
  | "position" when in_meta ->   check_arity_context_linetype s; TPosition
  | "comments" when in_meta ->   check_arity_context_linetype s; TComments
  | "format" when in_meta ->     check_arity_context_linetype s; TFormat
  | "analysis" when in_meta ->   check_arity_context_linetype s; TAnalysis
  | "any" when in_meta ->        check_arity_context_linetype s; TPosAny
  | "pure" when in_meta && in_iso ->
      check_arity_context_linetype s; TPure
  | "context" when in_meta && in_iso ->
      check_arity_context_linetype s; TContext
  | "error" when in_meta ->      check_arity_context_linetype s; TError
  | "words" when in_meta ->      check_context_linetype s; TWords
  | "symbol" when in_meta ->     check_arity_context_linetype s; TSymbol

  | "using" when in_rule_name || in_prolog ->  check_context_linetype s; TUsing
  | "virtual" when in_prolog || in_rule_name || in_meta ->
      (* don't want to allow virtual as a rule name *)
      check_context_linetype s; TVirtual
  | "merge" when in_prolog || in_rule_name || in_meta ->
      check_context_linetype s; TMerge
  | "disable" when in_rule_name ->  check_context_linetype s; TDisable
  | "extends" when in_rule_name -> check_context_linetype s; TExtends
  | "depends" when in_rule_name -> check_context_linetype s; TDepends
  | "on" when in_rule_name      -> check_context_linetype s; TOn
  | "ever" when in_rule_name    -> check_context_linetype s; TEver
  | "never" when in_rule_name   -> check_context_linetype s; TNever
  | "file" when in_rule_name    -> check_context_linetype s; TFile
  | "in" when in_rule_name      -> check_context_linetype s; TIn
  (* exists and forall for when are reparsed in parse_cocci.ml *)
  | "exists" when in_rule_name  -> check_context_linetype s; TExists
  | "forall" when in_rule_name  -> check_context_linetype s; TForall
  | "script" when (in_rule_name || in_meta) ->
      check_context_linetype s; TScript (!file, !line)
  | "initialize" when in_rule_name -> check_context_linetype s; TInitialize
  | "finalize" when in_rule_name   -> check_context_linetype s; TFinalize

  | "char" ->       Tchar     linetype
  | "short" ->      Tshort    linetype
  | "int" ->        Tint      linetype
  | "double" ->     Tdouble   linetype
  | "float" ->      Tfloat    linetype
  | "complex" ->    Tcomplex  linetype
  | "long" ->       Tlong     linetype
  | "void" ->       Tvoid     linetype
  | "size_t" ->     Tsize_t   linetype
  | "ssize_t" ->    Tssize_t  linetype
  | "ptrdiff_t" ->  Tptrdiff_t linetype
  (* in_meta is only for the first keyword; drop it now to allow any type
     name *)
  | "struct" ->     Data.saw_struct := true; Tstruct   linetype
  | "union" ->      Data.saw_struct := true; Tunion    linetype
  | "enum" ->       Data.saw_struct := true; Tenum     linetype
  | "unsigned" ->   Tunsigned linetype
  | "signed" ->     Tsigned   linetype
  | "decimal" when !Flag.ibm -> Tdecimal linetype
  | "EXEC" when !Flag.ibm -> Texec linetype

  | "auto"  ->      Tauto     linetype
  | "register" ->   Tregister linetype
  | "extern" ->     Textern   linetype
  | "static" ->     Tstatic   linetype
  | "inline" ->     Tinline   linetype
  | "typedef" ->    Ttypedef  linetype

  | "const" ->      Tconst    linetype
  | "volatile" ->   Tvolatile linetype

  | "if" ->         TIf       linetype
  | "else" ->       TElse     linetype
  | "while" ->      TWhile    linetype
  | "do" ->         TDo       linetype
  | "for" ->        TFor      linetype
  | "switch" ->     TSwitch   linetype
  | "case" ->       TCase     linetype
  | "default" ->    TDefault  linetype
  | "return" ->     TReturn   linetype
  | "break" ->      TBreak    linetype
  | "continue" ->   TContinue linetype
  | "goto" ->       TGoto     linetype

  | "sizeof" ->     TSizeof   linetype
  | "typeof" ->     TTypeof   linetype

  | "Expression"       when !Data.in_iso -> TIsoExpression
  | "ArgExpression"    when !Data.in_iso -> TIsoArgExpression
  | "TestExpression"   when !Data.in_iso -> TIsoTestExpression
  | "ToTestExpression" when !Data.in_iso -> TIsoToTestExpression
  | "Statement"        when !Data.in_iso -> TIsoStatement
  | "Declaration"      when !Data.in_iso -> TIsoDeclaration
  | "Type"             when !Data.in_iso -> TIsoType
  | "TopLevel"         when !Data.in_iso -> TIsoTopLevel

  | "_" when !Data.in_meta -> TUnderscore

  | s -> check_var s linetype

let mkassign op lexbuf =
  TOpAssign (op, (get_current_line_type lexbuf))

let spinit _ = (* per semantic patch *)
  Hashtbl.clear Data.non_local_script_constraints

let init _ = (* per file, first .cocci then iso *)
  line := 1;
  logical_line := 0;
  prev_plus := false;
  line_start := 0;
  current_line_started := false;
  current_line_type := (D.CONTEXT,0,0);
  col_zero := true;
  pm := UNKNOWN;
  Data.in_rule_name := false;
  Data.in_meta := false;
  Data.in_prolog := false;
  Data.saw_struct := false;
  Data.inheritable_positions := [];
  Hashtbl.clear all_metavariables;
  Hashtbl.clear Data.all_metadecls;
  Hashtbl.clear metavariables;
  Hashtbl.clear type_names;
  Hashtbl.clear rule_names;
  Hashtbl.clear iterator_names;
  Hashtbl.clear declarer_names;
  Hashtbl.clear symbol_names;
  let get_name (_,x) = x in
  Data.add_meta_meta :=
    (fun name cstr pure ->
      let fn clt = TMeta(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_id_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaId(name,constraints,Ast.NoVal,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_virt_id_meta_found :=
    (fun name vl ->
      let fn clt = TIdent(vl,clt) in
      Hashtbl.replace metavariables name fn);
  Data.add_virt_id_meta_not_found :=
    (fun name pure ->
      let fn clt = TMetaId(name,Ast.CstrTrue,Ast.NoVal,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_fresh_id_meta :=
    (fun name seed ->
      let fn clt = TMetaId(name,Ast.CstrTrue,seed,Ast0.Impure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_type_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaType(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_init_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaInit(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_initlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaInitList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_param_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaParam(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_paramlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaParamList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_const_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaConst(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_err_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaErr(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_exp_meta :=
    (fun tyopt name constraints pure bitfield ->
      let fn clt = TMetaExp(name,constraints,pure,tyopt,clt,bitfield) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_idexp_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaIdExp(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_local_idexp_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaLocalIdExp(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_global_idexp_meta :=
    (fun tyopt name constraints pure ->
      let fn clt = TMetaGlobalIdExp(name,constraints,pure,tyopt,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_explist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaExpList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_decl_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaDecl(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_field_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaField(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_field_list_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaFieldList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_fmt_meta :=
    (function name -> function iconstraints ->
      let fn clt = failwith "format metavariable only allowed in a string" in
      Data.format_metavariables :=
	(get_name name,(name,iconstraints)) :: !Data.format_metavariables;
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_fmtlist_meta :=
    (fun name constraints lenname ->
      let fn clt =
	failwith "format list metavariable only allowed in a string" in
      Data.format_list_metavariables :=
	(get_name name,(name,lenname,constraints))
	:: !Data.format_list_metavariables;
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_stm_meta :=
    (fun name cstr pure ->
      let fn clt = TMetaStm(name,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_stmlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaStmList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_dparamlist_meta :=
    (fun name lenname cstr pure ->
      let fn clt = TMetaDParamList(name,lenname,cstr,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_func_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaFunc(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_local_func_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaLocalFunc(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_iterator_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaIterator(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_declarer_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaDeclarer(name,constraints,pure,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_pos_meta :=
    (fun name constraints any ->
      let fn ((d,ln,_,_,_,_,_,_,_,_) as clt) =
	(if d = Data.PLUS
	then
	  failwith
	    (Printf.sprintf "%d: positions only allowed in minus code" ln));
	TMetaPos(name,constraints,any,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_com_meta :=
    (fun name constraints ->
      let fn ((d,ln,_,_,_,_,_,_,_,_) as clt) =
	(if d = Data.PLUS
	then
	  failwith
	    (Printf.sprintf "%d: comment variables only allowed in minus code" ln));
	TMetaCom(name,constraints,clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_assignOp_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaAssignOp (name, constraints, pure, clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_binaryOp_meta :=
    (fun name constraints pure ->
      let fn clt = TMetaBinaryOp (name, constraints, pure, clt) in
      Hashtbl.replace metavariables (get_name name) fn);
  Data.add_type_name :=
    (function name ->
      let fn clt = TTypeId(name,clt) in
      Hashtbl.replace type_names name fn);
  Data.add_attribute :=
    (function name ->
      let fn clt =
	match clt with
	  ((Data.PLUS | Data.PLUSPLUS),_,_,_,_,_,_,_,_,_) ->
	    TDirective (Ast.Space name, clt)
	| _ -> Tattr (name, clt) in
      Hashtbl.replace attr_names name fn);
  Data.add_declarer_name :=
    (function name ->
      let fn clt = TDeclarerId(name,clt) in
      Hashtbl.replace declarer_names name fn);
  Data.add_iterator_name :=
    (function name ->
      let fn clt = TIteratorId(name,clt) in
      Hashtbl.replace iterator_names name fn);
  Data.add_symbol_meta :=
    (function name ->
      let fn clt = TSymId (name,clt) in
      Hashtbl.replace symbol_names name fn);
  Data.init_rule := (function _ -> Hashtbl.clear metavariables);
  Data.install_bindings :=
    (function parent ->
      List.iter (function (name,fn) -> Hashtbl.add metavariables name fn)
	(Hashtbl.find all_metavariables parent))

let post_init _ =
  Stdcompat.Hashtbl.reset all_metavariables;
  Stdcompat.Hashtbl.reset Data.all_metadecls;
  Stdcompat.Hashtbl.reset metavariables;
  Stdcompat.Hashtbl.reset type_names;
  Stdcompat.Hashtbl.reset rule_names;
  Stdcompat.Hashtbl.reset iterator_names;
  Stdcompat.Hashtbl.reset declarer_names;
  Stdcompat.Hashtbl.reset symbol_names

(* initialization for each cocci rule *)
let reinit _ =
  Data.format_metavariables := [];
  Data.format_list_metavariables := []

(* the following is needed to properly tokenize include files.  Because an
include file is included after seeing a @, so current_line_started is true.
Current_line_started is not important for parsing the name of a rule, so we
don't have to reset this value to true after parsing an included file. *)
let include_init _ =
  current_line_started := false

let drop_spaces s =
  let len = String.length s in
  let rec loop n =
    if n = len
    then n
    else
      if List.mem (String.get s n) [' ';'\t']
      then loop (n+1)
      else n in
  let start = loop 0 in
  String.sub s start (len - start)

# 571 "parsing_cocci/lexer_cocci.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base =
   "\000\000\160\255\161\255\081\000\119\000\167\255\168\255\192\000\
    \023\001\038\000\067\000\101\000\079\000\218\255\081\000\082\000\
    \085\000\224\255\226\255\229\255\230\255\231\255\232\255\234\255\
    \086\000\192\000\238\255\087\000\240\255\116\000\118\000\140\000\
    \100\001\111\001\198\001\087\000\029\002\058\002\150\000\255\255\
    \144\002\119\000\106\000\174\000\175\000\177\255\211\255\049\000\
    \050\000\089\000\041\000\055\000\051\000\060\002\188\255\078\000\
    \089\000\091\000\185\000\153\002\228\002\094\000\101\000\098\000\
    \108\000\203\000\086\003\169\003\185\255\116\000\111\000\106\000\
    \119\000\222\000\027\004\141\004\177\002\241\000\064\002\119\000\
    \113\000\105\000\123\000\123\000\087\003\179\000\172\000\203\000\
    \088\003\204\000\183\255\001\001\182\255\180\000\181\255\243\000\
    \192\001\072\001\065\002\245\000\110\002\111\002\076\001\115\000\
    \183\000\066\002\182\000\184\000\195\001\197\001\181\000\186\000\
    \249\001\206\000\204\000\063\002\210\000\220\000\112\002\008\001\
    \199\004\030\005\117\005\204\005\035\006\122\006\209\006\040\007\
    \127\007\214\007\045\008\132\008\219\008\050\009\009\001\137\009\
    \224\009\026\010\113\010\171\010\010\001\199\000\002\011\012\001\
    \026\001\089\011\176\011\027\001\007\012\094\012\152\012\239\012\
    \250\255\070\013\157\013\244\013\075\014\162\014\245\003\041\001\
    \137\002\068\001\053\001\245\255\243\255\070\001\225\255\087\003\
    \050\003\255\003\055\001\058\001\097\001\124\001\189\001\190\001\
    \246\255\191\001\192\001\244\255\205\255\204\255\163\255\214\255\
    \194\001\228\255\113\002\213\255\197\001\115\002\200\255\004\000\
    \178\255\190\255\233\255\235\255\237\255\194\255\208\255\212\255\
    \199\255\181\001\183\001\206\255\203\255\210\255\162\255\201\255\
    \195\255\209\255\207\255\220\014\051\015\202\001\109\015\196\015\
    \203\001\027\016\114\016\172\016\003\017\164\255\093\002\186\001\
    \082\017\017\004\108\004\131\004\109\017\027\004\106\005\179\017\
    \214\001\165\255\067\002\017\018\196\255\097\018\135\018\200\255\
    \201\255\209\018\040\019\254\001\233\001\007\002\121\002\135\002\
    \009\002\027\002\127\002\128\002\133\002\239\255\240\255\193\005\
    \141\002\243\255\244\255\245\255\246\255\247\255\248\255\005\003\
    \053\003\126\002\168\003\255\255\158\017\151\002\119\003\209\255\
    \216\255\252\255\236\255\251\255\235\255\250\255\219\255\242\255\
    \025\006\132\009\111\006\198\006\234\255\139\002\211\255\217\255\
    \218\255\215\255\225\255\143\002\210\255\214\255\222\255\213\255\
    \221\255\212\255\098\019\185\019\147\002\243\019\074\020\148\002\
    \161\020\248\020\050\021\137\021\208\255\149\002\224\021\055\022\
    \150\002\142\022\229\022\031\023\118\023\176\023\154\002\088\002\
    \007\024\157\002\158\002\094\024\181\024\159\002\012\025\099\025\
    \157\025\244\025\197\255\069\017\216\002\067\026\030\007\084\011\
    \255\011\094\026\116\007\203\007\164\026\041\003\198\255\088\003\
    \255\255\098\011\252\255\234\026\022\005\034\008\254\255\001\027\
    \253\255\225\016\254\255\249\017\255\255\251\255\039\027\121\008\
    \208\008\253\255\062\027\252\255\162\017\253\255\233\021\254\255\
    \255\255\250\255\100\027\039\009\142\009\252\255\123\027\251\255\
    \091\007\252\255\253\255\254\255\197\002\255\255";
  Lexing.lex_backtrk =
   "\255\255\255\255\255\255\090\000\090\000\255\255\255\255\082\000\
    \081\000\053\000\066\000\064\000\040\000\255\255\036\000\059\000\
    \032\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \019\000\095\000\255\255\016\000\255\255\014\000\013\000\058\000\
    \028\000\081\000\081\000\006\000\081\000\095\000\033\000\255\255\
    \001\000\255\255\255\255\002\000\003\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\068\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\069\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\071\000\071\000\076\000\255\255\
    \255\255\255\255\255\255\255\255\075\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\255\255\
    \255\255\255\255\255\255\255\255\076\000\076\000\255\255\255\255\
    \076\000\255\255\255\255\076\000\255\255\255\255\076\000\255\255\
    \255\255\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\004\000\255\255\084\000\255\255\
    \085\000\255\255\083\000\255\255\255\255\255\255\255\255\083\000\
    \255\255\255\255\255\255\085\000\255\255\085\000\255\255\255\255\
    \255\255\081\000\081\000\007\000\081\000\081\000\089\000\255\255\
    \008\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \089\000\255\255\057\000\063\000\034\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \039\000\255\255\079\000\255\255\038\000\080\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\062\000\035\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\086\000\255\255\255\255\255\255\
    \086\000\255\255\086\000\255\255\255\255\255\255\090\000\090\000\
    \255\255\089\000\255\255\255\255\090\000\089\000\255\255\090\000\
    \090\000\255\255\090\000\255\255\255\255\057\000\057\000\255\255\
    \255\255\049\000\048\000\059\000\035\000\029\000\028\000\027\000\
    \032\000\025\000\022\000\018\000\031\000\255\255\255\255\014\000\
    \023\000\255\255\255\255\255\255\255\255\255\255\255\255\059\000\
    \006\000\059\000\024\000\255\255\001\000\255\255\002\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \056\000\255\255\056\000\255\255\255\255\017\000\255\255\255\255\
    \255\255\255\255\255\255\026\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\053\000\255\255\255\255\255\255\053\000\
    \255\255\053\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \051\000\255\255\052\000\255\255\050\000\255\255\255\255\255\255\
    \255\255\050\000\255\255\255\255\255\255\052\000\255\255\052\000\
    \255\255\255\255\255\255\057\000\057\000\255\255\056\000\255\255\
    \255\255\057\000\056\000\255\255\057\000\057\000\255\255\255\255\
    \255\255\000\000\255\255\003\000\001\000\001\000\255\255\002\000\
    \255\255\255\255\255\255\001\000\255\255\255\255\004\000\002\000\
    \002\000\255\255\003\000\255\255\255\255\255\255\002\000\255\255\
    \255\255\255\255\005\000\003\000\003\000\255\255\004\000\255\255\
    \255\255\255\255\255\255\255\255\004\000\255\255";
  Lexing.lex_default =
   "\001\000\000\000\000\000\255\255\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\000\000\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\043\000\044\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\077\000\077\000\095\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\091\000\089\000\
    \255\255\089\000\000\000\091\000\000\000\255\255\000\000\095\000\
    \095\000\095\000\095\000\099\000\095\000\095\000\102\000\255\255\
    \255\255\255\255\255\255\255\255\108\000\109\000\255\255\255\255\
    \112\000\255\255\255\255\115\000\255\255\255\255\118\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\000\000\000\000\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \255\255\000\000\186\000\000\000\255\255\189\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\236\000\000\000\255\255\255\255\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \255\255\255\255\255\255\000\000\255\255\255\255\014\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\000\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\088\001\
    \000\000\090\001\000\000\255\255\255\255\255\255\000\000\255\255\
    \000\000\098\001\000\000\101\001\000\000\000\000\255\255\255\255\
    \255\255\000\000\255\255\000\000\109\001\000\000\113\001\000\000\
    \000\000\000\000\255\255\255\255\255\255\000\000\255\255\000\000\
    \121\001\000\000\000\000\000\000\255\255\000\000";
  Lexing.lex_trans =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\040\000\039\000\039\000\039\000\039\000\192\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \040\000\027\000\005\000\037\000\007\000\014\000\011\000\006\000\
    \026\000\023\000\016\000\029\000\018\000\030\000\032\000\038\000\
    \004\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\009\000\017\000\031\000\012\000\015\000\028\000\
    \035\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\034\000\
    \008\000\008\000\008\000\022\000\025\000\021\000\010\000\036\000\
    \211\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\033\000\
    \008\000\008\000\008\000\020\000\024\000\019\000\013\000\225\000\
    \210\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\208\000\207\000\206\000\205\000\200\000\
    \201\000\202\000\199\000\198\000\190\000\221\000\226\000\152\000\
    \044\000\043\000\116\000\069\000\061\000\222\000\105\000\188\000\
    \104\000\055\000\209\000\184\000\103\000\225\000\223\000\228\000\
    \228\000\228\000\228\000\228\000\228\000\228\000\228\000\227\000\
    \227\000\187\000\056\000\183\000\185\000\221\000\226\000\173\000\
    \255\255\255\255\174\000\221\000\226\000\222\000\057\000\078\000\
    \045\000\058\000\059\000\222\000\062\000\043\000\223\000\079\000\
    \171\000\170\000\192\000\172\000\223\000\191\000\063\000\224\000\
    \064\000\065\000\197\000\046\000\066\000\070\000\071\000\072\000\
    \073\000\059\000\080\000\221\000\226\000\081\000\082\000\083\000\
    \084\000\093\000\094\000\222\000\007\000\113\000\193\000\074\000\
    \196\000\194\000\255\255\066\000\223\000\255\255\090\000\224\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\119\000\255\255\120\000\255\255\074\000\255\255\
    \002\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\110\000\109\000\108\000\111\000\007\000\
    \112\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\195\000\114\000\115\000\092\000\
    \117\000\118\000\135\000\119\000\141\000\137\000\140\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\119\000\255\255\120\000\148\000\144\000\255\255\160\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\165\000\164\000\166\000\182\000\008\000\181\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\159\000\007\000\158\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\180\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\119\000\177\000\120\000\255\255\098\000\255\255\255\255\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\255\255\255\255\255\255\255\255\008\000\255\255\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\156\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\007\000\175\000\176\000\178\000\179\000\186\000\
    \189\000\255\255\204\000\255\255\203\000\255\255\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \119\000\255\255\120\000\255\255\217\000\213\000\232\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\153\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\052\001\233\000\042\001\100\000\008\000\232\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\007\000\233\000\053\000\041\001\053\000\034\001\035\001\
    \255\255\255\255\255\255\255\255\255\255\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\119\000\
    \033\001\120\000\053\000\044\000\053\000\054\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \255\255\255\255\255\255\255\255\121\000\255\255\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \233\000\040\000\039\000\039\000\039\000\039\000\051\000\048\000\
    \051\000\048\000\059\000\049\000\097\000\049\000\047\000\099\000\
    \047\000\234\000\050\000\106\000\050\000\038\001\096\000\052\000\
    \040\000\052\000\233\000\041\000\162\000\107\000\039\001\161\000\
    \233\000\059\000\076\000\255\255\032\001\031\001\017\001\042\000\
    \255\255\029\001\028\001\255\255\037\001\255\255\014\001\163\000\
    \030\001\234\000\022\001\023\001\036\001\048\001\044\001\057\001\
    \053\001\076\000\233\000\101\000\063\001\102\000\059\001\062\001\
    \070\001\066\001\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\125\001\040\001\000\000\000\000\
    \060\000\255\255\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\085\001\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\255\255\
    \255\255\255\255\020\001\060\000\085\001\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\066\000\
    \088\000\088\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\018\001\255\255\000\000\086\001\066\000\088\000\
    \088\000\087\000\087\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\169\000\021\001\169\000\085\000\085\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\000\000\000\000\086\000\086\000\086\001\000\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\255\255\000\000\019\001\089\001\067\000\000\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\068\000\015\001\000\000\000\000\000\000\000\000\014\001\
    \000\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\000\000\000\000\016\001\000\000\000\000\
    \000\000\000\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\000\000\000\000\000\000\000\000\
    \067\000\000\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\074\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\000\000\167\000\074\000\000\000\000\000\000\000\000\000\
    \000\000\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\229\000\229\000\229\000\229\000\229\000\
    \229\000\229\000\229\000\229\000\229\000\000\000\167\000\000\000\
    \255\255\000\000\167\000\000\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\167\000\255\255\
    \000\000\000\000\075\000\000\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\076\000\230\000\
    \000\000\230\000\000\000\000\000\229\000\229\000\229\000\229\000\
    \229\000\229\000\229\000\229\000\229\000\229\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\076\000\000\000\000\000\
    \000\000\225\000\000\000\227\000\227\000\227\000\227\000\227\000\
    \227\000\227\000\227\000\227\000\227\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\000\000\
    \226\000\000\000\000\000\000\000\000\000\000\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \226\000\000\000\000\000\133\000\075\000\000\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\000\000\000\000\000\000\000\000\133\000\000\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\007\000\000\000\000\000\133\000\093\001\093\001\
    \093\001\093\001\093\001\093\001\093\001\093\001\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \119\000\000\000\120\000\000\000\000\000\000\000\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\000\000\000\000\000\000\000\000\008\000\000\000\122\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\007\000\229\000\229\000\229\000\229\000\229\000\229\000\
    \229\000\229\000\229\000\229\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\119\000\
    \000\000\120\000\000\000\000\000\000\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \000\000\000\000\000\000\000\000\008\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\123\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \007\000\024\001\024\001\024\001\024\001\024\001\024\001\024\001\
    \024\001\024\001\024\001\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\119\000\000\000\
    \120\000\000\000\000\000\000\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\000\000\
    \000\000\000\000\000\000\008\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \124\000\008\000\008\000\008\000\008\000\008\000\008\000\007\000\
    \000\000\024\001\024\001\024\001\024\001\024\001\024\001\024\001\
    \024\001\024\001\024\001\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\119\000\025\001\120\000\
    \000\000\000\000\000\000\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\025\001\000\000\
    \000\000\000\000\008\000\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\125\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\007\000\026\001\
    \026\001\026\001\026\001\026\001\026\001\026\001\026\001\026\001\
    \026\001\000\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\119\000\000\000\120\000\000\000\
    \000\000\000\000\000\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\000\000\000\000\000\000\
    \000\000\008\000\000\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\126\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\007\000\026\001\026\001\
    \026\001\026\001\026\001\026\001\026\001\026\001\026\001\026\001\
    \000\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\119\000\000\000\120\000\000\000\000\000\
    \000\000\000\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\000\000\000\000\000\000\000\000\
    \008\000\000\000\008\000\127\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\007\000\000\000\078\001\078\001\
    \078\001\078\001\078\001\078\001\078\001\078\001\078\001\078\001\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\119\000\025\001\120\000\123\001\123\001\123\001\
    \123\001\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\025\001\000\000\124\001\122\001\008\000\
    \000\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\128\000\008\000\008\000\
    \008\000\008\000\008\000\007\000\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\119\000\000\000\120\000\000\000\000\000\000\000\000\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\000\000\000\000\000\000\000\000\008\000\000\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\129\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\007\000\082\001\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \119\000\000\000\120\000\000\000\000\000\000\000\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\000\000\000\000\000\000\000\000\008\000\000\000\008\000\
    \008\000\008\000\008\000\130\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\007\000\094\001\094\001\094\001\094\001\094\001\094\001\
    \094\001\094\001\000\000\255\255\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\119\000\
    \000\000\120\000\000\000\000\000\000\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \000\000\000\000\000\000\000\000\131\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \007\000\104\001\104\001\104\001\104\001\104\001\104\001\104\001\
    \104\001\000\000\000\000\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\119\000\000\000\
    \120\000\000\000\000\000\000\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\000\000\
    \000\000\000\000\000\000\132\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\007\000\
    \105\001\105\001\105\001\105\001\105\001\105\001\105\001\105\001\
    \000\000\000\000\000\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\119\000\000\000\120\000\
    \000\000\000\000\000\000\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\000\000\000\000\
    \000\000\000\000\008\000\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\133\000\116\001\
    \116\001\116\001\116\001\116\001\116\001\116\001\116\001\000\000\
    \000\000\000\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\000\000\000\000\000\000\000\000\
    \134\000\000\000\000\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\000\000\000\000\000\000\
    \000\000\133\000\000\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\136\000\000\000\027\001\
    \133\000\027\001\000\000\000\000\026\001\026\001\026\001\026\001\
    \026\001\026\001\026\001\026\001\026\001\026\001\117\001\117\001\
    \117\001\117\001\117\001\117\001\117\001\117\001\000\000\000\000\
    \000\000\000\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\000\000\000\000\000\000\000\000\
    \136\000\000\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\000\000\000\000\137\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\144\000\000\000\145\000\000\000\000\000\000\000\
    \000\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\000\000\000\000\000\000\138\000\136\000\
    \000\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\000\000\000\000\000\000\
    \000\000\138\000\000\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\140\000\000\000\139\000\000\000\000\000\
    \000\000\000\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\000\000\000\000\000\000\142\000\
    \138\000\000\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\000\000\000\000\
    \000\000\000\000\142\000\000\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\000\000\
    \000\000\142\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\000\000\000\000\000\000\000\000\
    \143\000\000\000\000\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\000\000\000\000\000\000\
    \000\000\142\000\000\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\146\000\000\000\083\001\
    \142\000\083\001\000\000\000\000\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\000\000\000\000\
    \000\000\000\000\092\001\092\001\092\001\092\001\092\001\092\001\
    \092\001\092\001\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\000\000\000\000\000\000\000\000\
    \146\000\000\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\000\000\000\000\146\000\
    \000\000\000\000\091\001\000\000\000\000\000\000\000\000\000\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\000\000\000\000\000\000\000\000\147\000\000\000\
    \000\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\000\000\000\000\000\000\000\000\146\000\
    \000\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\149\000\000\000\078\001\146\000\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\079\001\000\000\000\000\000\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\255\255\000\000\079\001\000\000\149\000\000\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \144\000\000\000\150\000\000\000\000\000\000\000\000\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\000\000\000\000\000\000\151\000\149\000\000\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\000\000\000\000\000\000\000\000\151\000\
    \000\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\000\000\000\000\151\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\000\000\000\000\000\000\000\000\147\000\000\000\000\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\000\000\000\000\000\000\000\000\151\000\000\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\007\000\000\000\000\000\151\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \119\000\000\000\120\000\000\000\000\000\000\000\000\000\008\000\
    \008\000\008\000\008\000\154\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\000\000\000\000\000\000\000\000\008\000\000\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\007\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\119\000\
    \000\000\120\000\000\000\000\000\000\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\155\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \000\000\000\000\000\000\000\000\008\000\000\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \007\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\119\000\000\000\
    \120\000\000\000\000\000\000\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\000\000\
    \000\000\000\000\000\000\008\000\000\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\007\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\119\000\000\000\120\000\
    \000\000\000\000\000\000\000\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\000\000\000\000\
    \000\000\000\000\008\000\000\000\008\000\008\000\008\000\008\000\
    \157\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\007\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\119\000\000\000\120\000\000\000\
    \000\000\000\000\000\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\000\000\000\000\000\000\
    \212\000\008\000\000\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \155\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\000\000\
    \000\000\000\000\000\000\212\000\000\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\213\000\000\000\214\000\
    \000\000\000\000\000\000\000\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\000\000\000\000\
    \000\000\215\000\212\000\000\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \000\000\000\000\000\000\000\000\215\000\000\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\000\000\000\000\215\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\000\000\000\000\
    \000\000\000\000\216\000\000\000\000\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\000\000\
    \000\000\000\000\000\000\215\000\000\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\218\000\
    \000\000\000\000\215\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\000\000\000\000\
    \000\000\000\000\218\000\000\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\213\000\000\000\219\000\000\000\
    \000\000\000\000\000\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\000\000\000\000\000\000\
    \220\000\218\000\000\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\000\000\
    \100\001\000\000\000\000\220\000\000\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \000\000\000\000\220\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\099\001\000\000\000\000\
    \000\000\216\000\000\000\000\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\000\000\000\000\
    \000\000\000\000\220\000\000\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\000\000\000\000\
    \000\000\220\000\231\000\231\000\231\000\231\000\231\000\231\000\
    \231\000\231\000\231\000\231\000\000\000\000\000\000\000\000\000\
    \000\000\086\001\000\000\231\000\231\000\231\000\231\000\231\000\
    \231\000\000\000\086\001\225\000\000\000\228\000\228\000\228\000\
    \228\000\228\000\228\000\228\000\228\000\227\000\227\000\012\001\
    \011\001\011\001\011\001\011\001\111\001\111\001\111\001\111\001\
    \000\000\086\001\226\000\231\000\231\000\231\000\231\000\231\000\
    \231\000\222\000\086\001\000\000\000\000\000\000\012\001\000\000\
    \000\000\000\000\223\000\000\000\112\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\013\001\000\000\000\000\
    \000\000\000\000\226\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\222\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\223\000\231\000\231\000\231\000\231\000\231\000\
    \231\000\231\000\231\000\231\000\231\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\231\000\231\000\231\000\231\000\
    \231\000\231\000\000\000\000\000\000\000\000\000\110\001\222\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \223\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\231\000\231\000\231\000\231\000\
    \231\000\231\000\012\001\011\001\011\001\011\001\011\001\222\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \223\000\103\001\103\001\103\001\103\001\103\001\103\001\103\001\
    \103\001\012\001\007\001\239\000\243\000\241\000\249\000\247\000\
    \240\000\006\001\005\001\251\000\250\000\254\000\000\001\255\000\
    \010\001\238\000\237\000\237\000\237\000\237\000\237\000\237\000\
    \237\000\237\000\237\000\244\000\253\000\252\000\008\001\248\000\
    \000\000\009\001\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\004\001\000\000\003\001\245\000\
    \242\000\102\001\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\002\001\246\000\001\001\078\001\
    \000\000\237\000\237\000\237\000\237\000\237\000\237\000\237\000\
    \237\000\237\000\237\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\074\001\079\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\075\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\078\001\076\001\081\001\
    \081\001\081\001\081\001\081\001\081\001\081\001\081\001\080\001\
    \080\001\000\000\000\000\000\000\000\000\074\001\079\001\000\000\
    \000\000\000\000\000\000\074\001\079\001\075\001\000\000\000\000\
    \000\000\000\000\000\000\075\001\000\000\000\000\076\001\000\000\
    \000\000\000\000\000\000\000\000\076\001\000\000\000\000\077\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\074\001\079\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\075\001\000\000\241\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\076\001\000\000\000\000\077\001\
    \000\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\053\001\000\000\054\001\000\000\000\000\
    \000\000\255\255\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\000\000\000\000\000\000\000\000\
    \241\000\000\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\053\001\000\000\054\001\000\000\000\000\000\000\
    \000\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\000\000\000\000\000\000\043\001\242\000\
    \000\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\000\000\000\000\000\000\
    \000\000\043\001\000\000\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\044\001\000\000\045\001\000\000\000\000\
    \000\000\000\000\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\000\000\000\000\000\000\046\001\
    \043\001\000\000\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\000\000\000\000\
    \000\000\000\000\046\001\000\000\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\000\000\
    \000\000\046\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\000\000\000\000\000\000\000\000\
    \047\001\000\000\000\000\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\000\000\000\000\000\000\
    \000\000\046\001\000\000\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\049\001\000\000\000\000\
    \046\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\000\000\000\000\000\000\000\000\
    \049\001\000\000\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\044\001\000\000\050\001\000\000\000\000\000\000\
    \000\000\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\000\000\000\000\000\000\051\001\049\001\
    \000\000\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\000\000\000\000\000\000\
    \000\000\051\001\000\000\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\000\000\000\000\
    \051\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\000\000\000\000\000\000\000\000\047\001\
    \000\000\000\000\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\000\000\000\000\000\000\000\000\
    \051\001\000\000\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\055\001\000\000\000\000\051\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\115\001\115\001\115\001\115\001\115\001\115\001\115\001\
    \115\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\000\000\000\000\000\000\000\000\055\001\
    \000\000\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\000\000\000\000\055\001\000\000\
    \000\000\114\001\000\000\000\000\000\000\000\000\000\000\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\000\000\000\000\000\000\000\000\056\001\000\000\000\000\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\000\000\000\000\000\000\000\000\055\001\000\000\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\058\001\000\000\000\000\055\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\255\255\000\000\000\000\000\000\058\001\000\000\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\000\000\000\000\059\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\066\001\
    \000\000\067\001\000\000\000\000\000\000\000\000\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \000\000\000\000\000\000\060\001\058\001\000\000\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\000\000\000\000\000\000\000\000\060\001\000\000\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \062\001\000\000\061\001\000\000\000\000\000\000\000\000\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\000\000\000\000\000\000\064\001\060\001\000\000\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\000\000\000\000\000\000\000\000\064\001\
    \000\000\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\000\000\000\000\064\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\000\000\000\000\000\000\000\000\065\001\000\000\000\000\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\000\000\000\000\000\000\000\000\064\001\000\000\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\068\001\000\000\000\000\064\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\000\000\000\000\000\000\000\000\068\001\000\000\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\000\000\000\000\068\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\000\000\
    \000\000\000\000\000\000\069\001\000\000\000\000\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \000\000\000\000\000\000\000\000\068\001\000\000\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \071\001\000\000\000\000\068\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\000\000\
    \000\000\000\000\000\000\071\001\000\000\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\066\001\000\000\072\001\
    \000\000\000\000\000\000\000\000\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\000\000\000\000\
    \000\000\073\001\071\001\000\000\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \000\000\000\000\000\000\000\000\073\001\000\000\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\000\000\000\000\073\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\000\000\000\000\
    \000\000\000\000\069\001\000\000\000\000\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\000\000\
    \000\000\000\000\000\000\073\001\000\000\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\000\000\
    \000\000\000\000\073\001\084\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\084\001\084\001\084\001\084\001\
    \084\001\084\001\000\000\000\000\078\001\000\000\081\001\081\001\
    \081\001\081\001\081\001\081\001\081\001\081\001\080\001\080\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\079\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\075\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\076\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\079\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\075\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\076\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\084\001\084\001\084\001\
    \084\001\084\001\084\001\000\000\000\000\000\000\000\000\000\000\
    \075\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\076\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\084\001\084\001\084\001\
    \084\001\084\001\084\001\000\000\000\000\000\000\000\000\000\000\
    \075\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\076\001\095\001\095\001\095\001\095\001\095\001\095\001\
    \095\001\095\001\095\001\095\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\095\001\095\001\095\001\095\001\095\001\
    \095\001\096\001\096\001\096\001\096\001\096\001\096\001\096\001\
    \096\001\096\001\096\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\096\001\096\001\096\001\096\001\096\001\096\001\
    \000\000\000\000\000\000\095\001\095\001\095\001\095\001\095\001\
    \095\001\000\000\000\000\000\000\000\000\000\000\000\000\106\001\
    \106\001\106\001\106\001\106\001\106\001\106\001\106\001\106\001\
    \106\001\000\000\096\001\096\001\096\001\096\001\096\001\096\001\
    \106\001\106\001\106\001\106\001\106\001\106\001\107\001\107\001\
    \107\001\107\001\107\001\107\001\107\001\107\001\107\001\107\001\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\107\001\
    \107\001\107\001\107\001\107\001\107\001\000\000\000\000\000\000\
    \106\001\106\001\106\001\106\001\106\001\106\001\000\000\000\000\
    \000\000\000\000\000\000\000\000\118\001\118\001\118\001\118\001\
    \118\001\118\001\118\001\118\001\118\001\118\001\000\000\107\001\
    \107\001\107\001\107\001\107\001\107\001\118\001\118\001\118\001\
    \118\001\118\001\118\001\119\001\119\001\119\001\119\001\119\001\
    \119\001\119\001\119\001\119\001\119\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\119\001\119\001\119\001\119\001\
    \119\001\119\001\000\000\000\000\000\000\118\001\118\001\118\001\
    \118\001\118\001\118\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\119\001\119\001\119\001\119\001\
    \119\001\119\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000";
  Lexing.lex_check =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\000\000\191\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \009\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \010\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\011\000\012\000\012\000\014\000\015\000\
    \015\000\015\000\016\000\024\000\027\000\003\000\003\000\035\000\
    \041\000\042\000\047\000\050\000\051\000\003\000\048\000\029\000\
    \048\000\052\000\011\000\030\000\048\000\004\000\003\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\029\000\055\000\030\000\030\000\003\000\003\000\031\000\
    \043\000\044\000\031\000\004\000\004\000\003\000\056\000\049\000\
    \038\000\057\000\058\000\004\000\061\000\038\000\003\000\049\000\
    \031\000\031\000\025\000\031\000\004\000\025\000\062\000\004\000\
    \063\000\064\000\024\000\038\000\065\000\069\000\070\000\071\000\
    \072\000\058\000\079\000\004\000\004\000\080\000\081\000\082\000\
    \083\000\085\000\093\000\004\000\007\000\103\000\025\000\073\000\
    \025\000\025\000\086\000\065\000\004\000\087\000\089\000\004\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\077\000\007\000\095\000\073\000\099\000\
    \000\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\104\000\106\000\107\000\110\000\007\000\
    \111\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\008\000\025\000\113\000\114\000\091\000\
    \116\000\117\000\119\000\134\000\140\000\141\000\143\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\097\000\008\000\144\000\147\000\102\000\159\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\161\000\162\000\165\000\170\000\008\000\171\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\032\000\033\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\172\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\173\000\033\000\086\000\097\000\043\000\044\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\096\000\087\000\089\000\108\000\033\000\109\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\034\000\174\000\175\000\177\000\178\000\184\000\
    \188\000\077\000\201\000\095\000\202\000\099\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\091\000\034\000\112\000\213\000\216\000\223\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\243\000\232\000\244\000\096\000\034\000\223\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\034\000\034\000\034\000\034\000\034\000\034\000\034\000\
    \034\000\036\000\232\000\037\000\245\000\053\000\248\000\248\000\
    \097\000\115\000\078\000\098\000\102\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \249\000\036\000\037\000\037\000\053\000\037\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \100\000\101\000\118\000\186\000\036\000\189\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \036\000\036\000\036\000\036\000\036\000\036\000\036\000\036\000\
    \234\000\040\000\040\000\040\000\040\000\040\000\037\000\037\000\
    \053\000\053\000\059\000\037\000\078\000\053\000\037\000\098\000\
    \053\000\222\000\037\000\105\000\053\000\247\000\078\000\037\000\
    \040\000\053\000\222\000\040\000\160\000\105\000\246\000\160\000\
    \234\000\059\000\076\000\076\000\250\000\251\000\009\001\040\000\
    \096\000\252\000\252\000\108\000\247\000\109\000\013\001\160\000\
    \029\001\222\000\000\001\000\001\035\001\044\001\047\001\053\001\
    \056\001\076\000\222\000\100\000\062\001\101\000\063\001\065\001\
    \066\001\069\001\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\124\001\246\000\255\255\255\255\
    \059\000\112\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\076\001\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\115\000\
    \078\000\098\000\007\001\060\000\076\001\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\066\000\
    \084\000\088\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\255\255\255\255\100\000\101\000\
    \118\000\186\000\008\001\189\000\255\255\085\001\066\000\084\000\
    \088\000\084\000\088\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\014\001\167\000\007\001\167\000\084\000\088\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\255\255\255\255\084\000\088\000\085\001\255\255\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\076\000\255\255\008\001\087\001\066\000\255\255\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\067\000\010\001\255\255\255\255\255\255\255\255\010\001\
    \255\255\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\255\255\255\255\010\001\255\255\255\255\
    \255\255\255\255\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\255\255\255\255\255\255\255\255\
    \067\000\255\255\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\074\000\158\000\158\000\158\000\
    \158\000\158\000\158\000\158\000\158\000\158\000\158\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \169\000\255\255\158\000\074\000\255\255\255\255\255\255\255\255\
    \255\255\225\000\225\000\225\000\225\000\225\000\225\000\225\000\
    \225\000\225\000\225\000\229\000\229\000\229\000\229\000\229\000\
    \229\000\229\000\229\000\229\000\229\000\255\255\225\000\255\255\
    \087\001\255\255\158\000\255\255\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\225\000\014\001\
    \255\255\255\255\074\000\255\255\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\075\000\226\000\
    \255\255\226\000\255\255\255\255\226\000\226\000\226\000\226\000\
    \226\000\226\000\226\000\226\000\226\000\226\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\075\000\255\255\255\255\
    \255\255\227\000\255\255\227\000\227\000\227\000\227\000\227\000\
    \227\000\227\000\227\000\227\000\227\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\255\255\
    \227\000\255\255\255\255\255\255\255\255\255\255\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \227\000\255\255\255\255\120\000\075\000\255\255\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\255\255\255\255\255\255\255\255\120\000\255\255\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\121\000\255\255\255\255\120\000\092\001\092\001\
    \092\001\092\001\092\001\092\001\092\001\092\001\121\000\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\255\255\121\000\255\255\255\255\255\255\255\255\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\255\255\255\255\255\255\255\255\121\000\255\255\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\122\000\230\000\230\000\230\000\230\000\230\000\230\000\
    \230\000\230\000\230\000\230\000\255\255\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \255\255\122\000\255\255\255\255\255\255\255\255\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \255\255\255\255\255\255\255\255\122\000\255\255\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\122\000\122\000\
    \123\000\255\000\255\000\255\000\255\000\255\000\255\000\255\000\
    \255\000\255\000\255\000\255\255\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\255\255\
    \123\000\255\255\255\255\255\255\255\255\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\255\255\
    \255\255\255\255\255\255\123\000\255\255\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\123\000\124\000\
    \255\255\024\001\024\001\024\001\024\001\024\001\024\001\024\001\
    \024\001\024\001\024\001\124\000\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\024\001\124\000\
    \255\255\255\255\255\255\255\255\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\024\001\255\255\
    \255\255\255\255\124\000\255\255\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\124\000\125\000\026\001\
    \026\001\026\001\026\001\026\001\026\001\026\001\026\001\026\001\
    \026\001\255\255\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\255\255\125\000\255\255\
    \255\255\255\255\255\255\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\255\255\255\255\255\255\
    \255\255\125\000\255\255\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\125\000\125\000\125\000\
    \125\000\125\000\125\000\125\000\125\000\126\000\027\001\027\001\
    \027\001\027\001\027\001\027\001\027\001\027\001\027\001\027\001\
    \255\255\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\255\255\126\000\255\255\255\255\
    \255\255\255\255\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\255\255\255\255\255\255\255\255\
    \126\000\255\255\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\126\000\126\000\126\000\126\000\
    \126\000\126\000\126\000\126\000\127\000\255\255\078\001\078\001\
    \078\001\078\001\078\001\078\001\078\001\078\001\078\001\078\001\
    \127\000\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\078\001\127\000\120\001\120\001\120\001\
    \120\001\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\078\001\255\255\120\001\120\001\127\000\
    \255\255\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\127\000\127\000\127\000\127\000\127\000\
    \127\000\127\000\127\000\128\000\082\001\082\001\082\001\082\001\
    \082\001\082\001\082\001\082\001\082\001\082\001\255\255\128\000\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\255\255\128\000\255\255\255\255\255\255\255\255\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\255\255\255\255\255\255\255\255\128\000\255\255\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\128\000\128\000\128\000\128\000\128\000\128\000\
    \128\000\128\000\129\000\083\001\083\001\083\001\083\001\083\001\
    \083\001\083\001\083\001\083\001\083\001\255\255\129\000\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\
    \129\000\255\255\129\000\255\255\255\255\255\255\255\255\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\
    \129\000\255\255\255\255\255\255\255\255\129\000\255\255\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\
    \129\000\129\000\129\000\129\000\129\000\129\000\129\000\129\000\
    \129\000\130\000\093\001\093\001\093\001\093\001\093\001\093\001\
    \093\001\093\001\255\255\120\001\255\255\130\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \255\255\130\000\255\255\255\255\255\255\255\255\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \255\255\255\255\255\255\255\255\130\000\255\255\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \130\000\130\000\130\000\130\000\130\000\130\000\130\000\130\000\
    \131\000\103\001\103\001\103\001\103\001\103\001\103\001\103\001\
    \103\001\255\255\255\255\255\255\131\000\131\000\131\000\131\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\255\255\
    \131\000\255\255\255\255\255\255\255\255\131\000\131\000\131\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\131\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\131\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\255\255\
    \255\255\255\255\255\255\131\000\255\255\131\000\131\000\131\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\131\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\131\000\
    \131\000\131\000\131\000\131\000\131\000\131\000\131\000\132\000\
    \104\001\104\001\104\001\104\001\104\001\104\001\104\001\104\001\
    \255\255\255\255\255\255\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\255\255\132\000\
    \255\255\255\255\255\255\255\255\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\255\255\255\255\
    \255\255\255\255\132\000\255\255\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\132\000\132\000\
    \132\000\132\000\132\000\132\000\132\000\132\000\133\000\115\001\
    \115\001\115\001\115\001\115\001\115\001\115\001\115\001\255\255\
    \255\255\255\255\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\255\255\255\255\255\255\255\255\
    \133\000\255\255\255\255\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\255\255\255\255\255\255\
    \255\255\133\000\255\255\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\133\000\133\000\133\000\
    \133\000\133\000\133\000\133\000\133\000\135\000\255\255\025\001\
    \133\000\025\001\255\255\255\255\025\001\025\001\025\001\025\001\
    \025\001\025\001\025\001\025\001\025\001\025\001\116\001\116\001\
    \116\001\116\001\116\001\116\001\116\001\116\001\255\255\255\255\
    \255\255\255\255\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\255\255\255\255\255\255\255\255\
    \135\000\255\255\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\135\000\135\000\135\000\135\000\
    \135\000\135\000\135\000\135\000\136\000\255\255\255\255\135\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\255\255\136\000\255\255\255\255\255\255\
    \255\255\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\255\255\255\255\255\255\137\000\136\000\
    \255\255\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\136\000\136\000\136\000\136\000\136\000\
    \136\000\136\000\136\000\137\000\137\000\137\000\137\000\137\000\
    \137\000\137\000\137\000\137\000\137\000\137\000\137\000\137\000\
    \137\000\137\000\137\000\137\000\137\000\137\000\137\000\137\000\
    \137\000\137\000\137\000\137\000\137\000\255\255\255\255\255\255\
    \255\255\137\000\255\255\137\000\137\000\137\000\137\000\137\000\
    \137\000\137\000\137\000\137\000\137\000\137\000\137\000\137\000\
    \137\000\137\000\137\000\137\000\137\000\137\000\137\000\137\000\
    \137\000\137\000\137\000\137\000\137\000\138\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\255\255\138\000\255\255\255\255\
    \255\255\255\255\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\255\255\255\255\255\255\139\000\
    \138\000\255\255\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\138\000\138\000\138\000\138\000\
    \138\000\138\000\138\000\138\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\255\255\255\255\
    \255\255\255\255\139\000\255\255\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\139\000\139\000\
    \139\000\139\000\139\000\139\000\139\000\139\000\142\000\255\255\
    \255\255\139\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\255\255\255\255\255\255\255\255\
    \142\000\255\255\255\255\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\255\255\255\255\255\255\
    \255\255\142\000\255\255\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\142\000\142\000\142\000\
    \142\000\142\000\142\000\142\000\142\000\145\000\255\255\079\001\
    \142\000\079\001\255\255\255\255\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\079\001\079\001\255\255\255\255\
    \255\255\255\255\089\001\089\001\089\001\089\001\089\001\089\001\
    \089\001\089\001\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\255\255\255\255\255\255\255\255\
    \145\000\255\255\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\145\000\145\000\145\000\145\000\
    \145\000\145\000\145\000\145\000\146\000\255\255\255\255\145\000\
    \255\255\255\255\089\001\255\255\255\255\255\255\255\255\255\255\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\255\255\255\255\255\255\255\255\146\000\255\255\
    \255\255\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\255\255\255\255\255\255\255\255\146\000\
    \255\255\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\146\000\146\000\146\000\146\000\146\000\
    \146\000\146\000\146\000\148\000\255\255\080\001\146\000\080\001\
    \080\001\080\001\080\001\080\001\080\001\080\001\080\001\080\001\
    \080\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\080\001\255\255\255\255\255\255\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\089\001\255\255\080\001\255\255\148\000\255\255\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\148\000\148\000\148\000\148\000\148\000\148\000\
    \148\000\148\000\149\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\255\255\149\000\255\255\255\255\255\255\255\255\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\255\255\255\255\255\255\150\000\149\000\255\255\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\149\000\149\000\149\000\149\000\149\000\149\000\149\000\
    \149\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\255\255\255\255\255\255\255\255\150\000\
    \255\255\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\150\000\150\000\150\000\150\000\150\000\
    \150\000\150\000\150\000\151\000\255\255\255\255\150\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\255\255\255\255\255\255\255\255\151\000\255\255\255\255\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\255\255\255\255\255\255\255\255\151\000\255\255\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\151\000\151\000\151\000\151\000\151\000\151\000\
    \151\000\151\000\153\000\255\255\255\255\151\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\255\255\153\000\255\255\255\255\255\255\255\255\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\255\255\255\255\255\255\255\255\153\000\255\255\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\153\000\153\000\153\000\153\000\153\000\153\000\153\000\
    \153\000\154\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \255\255\154\000\255\255\255\255\255\255\255\255\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \255\255\255\255\255\255\255\255\154\000\255\255\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \155\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\255\255\
    \155\000\255\255\255\255\255\255\255\255\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\255\255\
    \255\255\255\255\255\255\155\000\255\255\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\155\000\
    \155\000\155\000\155\000\155\000\155\000\155\000\155\000\156\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\255\255\156\000\
    \255\255\255\255\255\255\255\255\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\255\255\255\255\
    \255\255\255\255\156\000\255\255\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\156\000\156\000\
    \156\000\156\000\156\000\156\000\156\000\156\000\157\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\157\000\157\000\157\000\157\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\255\255\157\000\255\255\
    \255\255\255\255\255\255\157\000\157\000\157\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\255\255\255\255\255\255\
    \211\000\157\000\255\255\157\000\157\000\157\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\157\000\157\000\157\000\
    \157\000\157\000\157\000\157\000\157\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\255\255\
    \255\255\255\255\255\255\211\000\255\255\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\211\000\
    \211\000\211\000\211\000\211\000\211\000\211\000\211\000\212\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\255\255\212\000\
    \255\255\255\255\255\255\255\255\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\255\255\255\255\
    \255\255\214\000\212\000\255\255\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\212\000\212\000\
    \212\000\212\000\212\000\212\000\212\000\212\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \255\255\255\255\255\255\255\255\214\000\255\255\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \214\000\214\000\214\000\214\000\214\000\214\000\214\000\214\000\
    \215\000\255\255\255\255\214\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\255\255\255\255\
    \255\255\255\255\215\000\255\255\255\255\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\255\255\
    \255\255\255\255\255\255\215\000\255\255\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\215\000\
    \215\000\215\000\215\000\215\000\215\000\215\000\215\000\217\000\
    \255\255\255\255\215\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\255\255\255\255\
    \255\255\255\255\217\000\255\255\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\217\000\217\000\
    \217\000\217\000\217\000\217\000\217\000\217\000\218\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\255\255\218\000\255\255\
    \255\255\255\255\255\255\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\255\255\255\255\255\255\
    \219\000\218\000\255\255\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\218\000\218\000\218\000\
    \218\000\218\000\218\000\218\000\218\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\255\255\
    \097\001\255\255\255\255\219\000\255\255\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\219\000\
    \219\000\219\000\219\000\219\000\219\000\219\000\219\000\220\000\
    \255\255\255\255\219\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\097\001\255\255\255\255\
    \255\255\220\000\255\255\255\255\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\255\255\255\255\
    \255\255\255\255\220\000\255\255\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\220\000\220\000\
    \220\000\220\000\220\000\220\000\220\000\220\000\255\255\255\255\
    \255\255\220\000\224\000\224\000\224\000\224\000\224\000\224\000\
    \224\000\224\000\224\000\224\000\255\255\255\255\255\255\255\255\
    \255\255\075\001\255\255\224\000\224\000\224\000\224\000\224\000\
    \224\000\255\255\075\001\228\000\255\255\228\000\228\000\228\000\
    \228\000\228\000\228\000\228\000\228\000\228\000\228\000\012\001\
    \012\001\012\001\012\001\012\001\108\001\108\001\108\001\108\001\
    \255\255\075\001\228\000\224\000\224\000\224\000\224\000\224\000\
    \224\000\228\000\075\001\255\255\255\255\255\255\012\001\255\255\
    \255\255\255\255\228\000\255\255\108\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\012\001\255\255\255\255\
    \255\255\255\255\228\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\228\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\097\001\228\000\231\000\231\000\231\000\231\000\231\000\
    \231\000\231\000\231\000\231\000\231\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\231\000\231\000\231\000\231\000\
    \231\000\231\000\255\255\255\255\255\255\255\255\108\001\231\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \231\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\231\000\231\000\231\000\231\000\
    \231\000\231\000\235\000\235\000\235\000\235\000\235\000\231\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \231\000\099\001\099\001\099\001\099\001\099\001\099\001\099\001\
    \099\001\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \255\255\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\255\255\235\000\235\000\
    \235\000\099\001\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\235\000\
    \235\000\235\000\235\000\235\000\235\000\235\000\235\000\237\000\
    \255\255\237\000\237\000\237\000\237\000\237\000\237\000\237\000\
    \237\000\237\000\237\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\108\001\255\255\255\255\237\000\237\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\237\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\238\000\237\000\238\000\
    \238\000\238\000\238\000\238\000\238\000\238\000\238\000\238\000\
    \238\000\255\255\255\255\255\255\255\255\237\000\237\000\255\255\
    \255\255\255\255\255\255\238\000\238\000\237\000\255\255\255\255\
    \255\255\255\255\255\255\238\000\255\255\255\255\237\000\255\255\
    \255\255\255\255\255\255\255\255\238\000\255\255\255\255\238\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\238\000\238\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\238\000\255\255\241\000\255\255\255\255\
    \255\255\099\001\255\255\255\255\238\000\255\255\255\255\238\000\
    \255\255\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\255\255\241\000\255\255\255\255\
    \255\255\235\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\255\255\255\255\255\255\255\255\
    \241\000\255\255\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\241\000\241\000\241\000\241\000\
    \241\000\241\000\241\000\241\000\242\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\255\255\242\000\255\255\255\255\255\255\
    \255\255\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\255\255\255\255\255\255\042\001\242\000\
    \255\255\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\242\000\242\000\242\000\242\000\242\000\
    \242\000\242\000\242\000\042\001\042\001\042\001\042\001\042\001\
    \042\001\042\001\042\001\042\001\042\001\042\001\042\001\042\001\
    \042\001\042\001\042\001\042\001\042\001\042\001\042\001\042\001\
    \042\001\042\001\042\001\042\001\042\001\255\255\255\255\255\255\
    \255\255\042\001\255\255\042\001\042\001\042\001\042\001\042\001\
    \042\001\042\001\042\001\042\001\042\001\042\001\042\001\042\001\
    \042\001\042\001\042\001\042\001\042\001\042\001\042\001\042\001\
    \042\001\042\001\042\001\042\001\042\001\043\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\255\255\043\001\255\255\255\255\
    \255\255\255\255\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\255\255\255\255\255\255\045\001\
    \043\001\255\255\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\043\001\043\001\043\001\043\001\
    \043\001\043\001\043\001\043\001\045\001\045\001\045\001\045\001\
    \045\001\045\001\045\001\045\001\045\001\045\001\045\001\045\001\
    \045\001\045\001\045\001\045\001\045\001\045\001\045\001\045\001\
    \045\001\045\001\045\001\045\001\045\001\045\001\255\255\255\255\
    \255\255\255\255\045\001\255\255\045\001\045\001\045\001\045\001\
    \045\001\045\001\045\001\045\001\045\001\045\001\045\001\045\001\
    \045\001\045\001\045\001\045\001\045\001\045\001\045\001\045\001\
    \045\001\045\001\045\001\045\001\045\001\045\001\046\001\255\255\
    \255\255\045\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\255\255\255\255\255\255\255\255\
    \046\001\255\255\255\255\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\255\255\255\255\255\255\
    \255\255\046\001\255\255\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\046\001\046\001\046\001\
    \046\001\046\001\046\001\046\001\046\001\048\001\255\255\255\255\
    \046\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\048\001\048\001\048\001\048\001\048\001\048\001\
    \048\001\048\001\048\001\048\001\048\001\048\001\048\001\048\001\
    \048\001\048\001\048\001\048\001\048\001\048\001\048\001\048\001\
    \048\001\048\001\048\001\048\001\255\255\255\255\255\255\255\255\
    \048\001\255\255\048\001\048\001\048\001\048\001\048\001\048\001\
    \048\001\048\001\048\001\048\001\048\001\048\001\048\001\048\001\
    \048\001\048\001\048\001\048\001\048\001\048\001\048\001\048\001\
    \048\001\048\001\048\001\048\001\049\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\255\255\049\001\255\255\255\255\255\255\
    \255\255\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\255\255\255\255\255\255\050\001\049\001\
    \255\255\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\049\001\049\001\049\001\049\001\049\001\
    \049\001\049\001\049\001\050\001\050\001\050\001\050\001\050\001\
    \050\001\050\001\050\001\050\001\050\001\050\001\050\001\050\001\
    \050\001\050\001\050\001\050\001\050\001\050\001\050\001\050\001\
    \050\001\050\001\050\001\050\001\050\001\255\255\255\255\255\255\
    \255\255\050\001\255\255\050\001\050\001\050\001\050\001\050\001\
    \050\001\050\001\050\001\050\001\050\001\050\001\050\001\050\001\
    \050\001\050\001\050\001\050\001\050\001\050\001\050\001\050\001\
    \050\001\050\001\050\001\050\001\050\001\051\001\255\255\255\255\
    \050\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\255\255\255\255\255\255\255\255\051\001\
    \255\255\255\255\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\255\255\255\255\255\255\255\255\
    \051\001\255\255\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\051\001\051\001\051\001\051\001\
    \051\001\051\001\051\001\051\001\054\001\255\255\255\255\051\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\110\001\110\001\110\001\110\001\110\001\110\001\110\001\
    \110\001\054\001\054\001\054\001\054\001\054\001\054\001\054\001\
    \054\001\054\001\054\001\054\001\054\001\054\001\054\001\054\001\
    \054\001\054\001\054\001\054\001\054\001\054\001\054\001\054\001\
    \054\001\054\001\054\001\255\255\255\255\255\255\255\255\054\001\
    \255\255\054\001\054\001\054\001\054\001\054\001\054\001\054\001\
    \054\001\054\001\054\001\054\001\054\001\054\001\054\001\054\001\
    \054\001\054\001\054\001\054\001\054\001\054\001\054\001\054\001\
    \054\001\054\001\054\001\055\001\255\255\255\255\054\001\255\255\
    \255\255\110\001\255\255\255\255\255\255\255\255\255\255\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\255\255\255\255\255\255\255\255\055\001\255\255\255\255\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\255\255\255\255\255\255\255\255\055\001\255\255\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\055\001\055\001\055\001\055\001\055\001\055\001\
    \055\001\055\001\057\001\255\255\255\255\055\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\057\001\
    \057\001\057\001\057\001\057\001\057\001\057\001\057\001\057\001\
    \057\001\057\001\057\001\057\001\057\001\057\001\057\001\057\001\
    \057\001\057\001\057\001\057\001\057\001\057\001\057\001\057\001\
    \057\001\110\001\255\255\255\255\255\255\057\001\255\255\057\001\
    \057\001\057\001\057\001\057\001\057\001\057\001\057\001\057\001\
    \057\001\057\001\057\001\057\001\057\001\057\001\057\001\057\001\
    \057\001\057\001\057\001\057\001\057\001\057\001\057\001\057\001\
    \057\001\058\001\255\255\255\255\057\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \255\255\058\001\255\255\255\255\255\255\255\255\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \255\255\255\255\255\255\059\001\058\001\255\255\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \058\001\058\001\058\001\058\001\058\001\058\001\058\001\058\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\255\255\255\255\255\255\255\255\059\001\255\255\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\059\001\059\001\059\001\059\001\059\001\059\001\
    \059\001\059\001\060\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\255\255\060\001\255\255\255\255\255\255\255\255\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\255\255\255\255\255\255\061\001\060\001\255\255\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\060\001\060\001\060\001\060\001\060\001\060\001\060\001\
    \060\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\255\255\255\255\255\255\255\255\061\001\
    \255\255\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\061\001\061\001\061\001\061\001\061\001\
    \061\001\061\001\061\001\064\001\255\255\255\255\061\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\255\255\255\255\255\255\255\255\064\001\255\255\255\255\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\255\255\255\255\255\255\255\255\064\001\255\255\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\064\001\064\001\064\001\064\001\064\001\064\001\
    \064\001\064\001\067\001\255\255\255\255\064\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\255\255\255\255\255\255\255\255\067\001\255\255\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\068\001\255\255\255\255\067\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\255\255\
    \255\255\255\255\255\255\068\001\255\255\255\255\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \255\255\255\255\255\255\255\255\068\001\255\255\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\068\001\068\001\
    \070\001\255\255\255\255\068\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\255\255\
    \255\255\255\255\255\255\070\001\255\255\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\070\001\
    \070\001\070\001\070\001\070\001\070\001\070\001\070\001\071\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\255\255\071\001\
    \255\255\255\255\255\255\255\255\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\255\255\255\255\
    \255\255\072\001\071\001\255\255\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\071\001\071\001\
    \071\001\071\001\071\001\071\001\071\001\071\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \255\255\255\255\255\255\255\255\072\001\255\255\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \072\001\072\001\072\001\072\001\072\001\072\001\072\001\072\001\
    \073\001\255\255\255\255\072\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\255\255\255\255\
    \255\255\255\255\073\001\255\255\255\255\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\255\255\
    \255\255\255\255\255\255\073\001\255\255\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\073\001\073\001\073\001\073\001\073\001\255\255\
    \255\255\255\255\073\001\077\001\077\001\077\001\077\001\077\001\
    \077\001\077\001\077\001\077\001\077\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\077\001\077\001\077\001\077\001\
    \077\001\077\001\255\255\255\255\081\001\255\255\081\001\081\001\
    \081\001\081\001\081\001\081\001\081\001\081\001\081\001\081\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\081\001\077\001\077\001\077\001\077\001\
    \077\001\077\001\081\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\081\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\081\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\081\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\081\001\084\001\084\001\084\001\084\001\
    \084\001\084\001\084\001\084\001\084\001\084\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\084\001\084\001\084\001\
    \084\001\084\001\084\001\255\255\255\255\255\255\255\255\255\255\
    \084\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\084\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\084\001\084\001\084\001\
    \084\001\084\001\084\001\255\255\255\255\255\255\255\255\255\255\
    \084\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\084\001\091\001\091\001\091\001\091\001\091\001\091\001\
    \091\001\091\001\091\001\091\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\091\001\091\001\091\001\091\001\091\001\
    \091\001\095\001\095\001\095\001\095\001\095\001\095\001\095\001\
    \095\001\095\001\095\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\095\001\095\001\095\001\095\001\095\001\095\001\
    \255\255\255\255\255\255\091\001\091\001\091\001\091\001\091\001\
    \091\001\255\255\255\255\255\255\255\255\255\255\255\255\102\001\
    \102\001\102\001\102\001\102\001\102\001\102\001\102\001\102\001\
    \102\001\255\255\095\001\095\001\095\001\095\001\095\001\095\001\
    \102\001\102\001\102\001\102\001\102\001\102\001\106\001\106\001\
    \106\001\106\001\106\001\106\001\106\001\106\001\106\001\106\001\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\106\001\
    \106\001\106\001\106\001\106\001\106\001\255\255\255\255\255\255\
    \102\001\102\001\102\001\102\001\102\001\102\001\255\255\255\255\
    \255\255\255\255\255\255\255\255\114\001\114\001\114\001\114\001\
    \114\001\114\001\114\001\114\001\114\001\114\001\255\255\106\001\
    \106\001\106\001\106\001\106\001\106\001\114\001\114\001\114\001\
    \114\001\114\001\114\001\118\001\118\001\118\001\118\001\118\001\
    \118\001\118\001\118\001\118\001\118\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\118\001\118\001\118\001\118\001\
    \118\001\118\001\255\255\255\255\255\255\114\001\114\001\114\001\
    \114\001\114\001\114\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\118\001\118\001\118\001\118\001\
    \118\001\118\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255";
  Lexing.lex_base_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\002\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\004\000\005\000\000\000\038\000\000\000\000\000\000\000\
    \000\000\006\000\038\000\152\000\007\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\006\000\009\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\018\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\033\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\055\000\055\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_default_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code =
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\001\000\015\000\015\000\028\000\028\000\046\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\001\000\015\000\015\000\028\000\028\000\046\000\052\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\046\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\046\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\043\000\000\000\000\000\000\000\023\000\012\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\000\000\000\000\000\000\000\000\049\000\000\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\052\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\000\000\000\000\000\000\000\000\049\000\
    \000\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\049\000\049\000\049\000\049\000\049\000\
    \049\000\049\000\049\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_check_code =
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\040\000\058\000\059\000\065\000\066\000\073\000\
    \076\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\040\000\058\000\059\000\065\000\066\000\073\000\076\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\074\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\074\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\072\000\255\255\255\255\255\255\064\000\057\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\255\255\255\255\255\255\255\255\074\000\255\255\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\075\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \075\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\255\255\255\255\255\255\255\255\075\000\
    \255\255\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255";
  Lexing.lex_code =
   "\255\005\255\004\255\255\000\004\255\000\005\255\006\255\255\007\
    \255\255\000\006\001\007\255\009\255\008\255\255\011\255\010\255\
    \255\000\008\001\010\255\000\009\001\011\255\012\255\255\013\255\
    \255\014\255\255\015\255\255\003\015\001\013\000\012\002\014\255\
    ";
}

let rec token lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 16 (-1);(* L=2 [5] <- p ; [4] <- p ;  *)
  lexbuf.Lexing.lex_mem.(5) <- lexbuf.Lexing.lex_curr_pos ;
  lexbuf.Lexing.lex_mem.(4) <- lexbuf.Lexing.lex_curr_pos ;
 __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 599 "parsing_cocci/lexer_cocci.mll"
    ( let cls = !current_line_started in
      if not cls
      then
	begin
	  match !current_line_type with
	    (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
	      (* increment the logical line even though nothing seen *)
	      start_line true;
	      let info = get_current_line_type lexbuf in
	      reset_line lexbuf;
	      TDirective (Ast.Noindent "", info)
	  | _ -> reset_line lexbuf; token lexbuf
	end
      else (reset_line lexbuf; token lexbuf) )
# 2830 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 614 "parsing_cocci/lexer_cocci.mll"
                      w
# 2836 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 614 "parsing_cocci/lexer_cocci.mll"
                         ( (* collect whitespaces only when inside a rule *)
    start_line false;
    if !Data.in_rule_name || !Data.in_prolog || !Data.in_iso
    then token lexbuf
    else TWhitespace w )
# 2844 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 620 "parsing_cocci/lexer_cocci.mll"
                                       after
# 2850 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 620 "parsing_cocci/lexer_cocci.mll"
                                              (
    match !current_line_type with
      (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
	let str =
	  if !current_line_started
	  then (tok lexbuf)
	  else after in
	start_line true;
	TDirective (Ast.Indent str, get_current_line_type lexbuf)
    | _ -> start_line false; token lexbuf )
# 2863 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 631 "parsing_cocci/lexer_cocci.mll"
                                       after
# 2869 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 631 "parsing_cocci/lexer_cocci.mll"
                                              (
     check_context_linetype after;
     start_line false; token lexbuf )
# 2875 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
# 636 "parsing_cocci/lexer_cocci.mll"
   ( match !current_line_type with
      (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
	start_line true;
	TAttr_ (get_current_line_type lexbuf)
    | _ -> failwith "attributes only allowed in + code" )
# 2884 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 642 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TArobArob )
# 2889 "parsing_cocci/lexer_cocci.ml"

  | 6 ->
# 643 "parsing_cocci/lexer_cocci.mll"
         ( pass_zero();
	   if !Data.in_rule_name || not !current_line_started
	   then (start_line true; TArob)
	   else (check_minus_context_linetype "@";
		 TPArob (get_current_line_type lexbuf)) )
# 2898 "parsing_cocci/lexer_cocci.ml"

  | 7 ->
# 650 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; check_minus_context_linetype (tok lexbuf);
	TWhen (get_current_line_type lexbuf) )
# 2904 "parsing_cocci/lexer_cocci.ml"

  | 8 ->
# 654 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; check_minus_context_linetype (tok lexbuf);
	TEllipsis (get_current_line_type lexbuf) )
# 2910 "parsing_cocci/lexer_cocci.ml"

  | 9 ->
# 656 "parsing_cocci/lexer_cocci.mll"
           ( start_line true; check_context_linetype (tok lexbuf);
	     TOEllipsis (get_current_line_type lexbuf) )
# 2916 "parsing_cocci/lexer_cocci.ml"

  | 10 ->
# 658 "parsing_cocci/lexer_cocci.mll"
           ( start_line true; check_context_linetype (tok lexbuf);
	     TCEllipsis (get_current_line_type lexbuf) )
# 2922 "parsing_cocci/lexer_cocci.ml"

  | 11 ->
# 660 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; check_minus_context_linetype (tok lexbuf);
	     TPOEllipsis (get_current_line_type lexbuf) )
# 2928 "parsing_cocci/lexer_cocci.ml"

  | 12 ->
# 662 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; check_minus_context_linetype (tok lexbuf);
	     TPCEllipsis (get_current_line_type lexbuf) )
# 2934 "parsing_cocci/lexer_cocci.ml"

  | 13 ->
# 664 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true; TMinus (get_current_line_type lexbuf))
          else (patch_or_match PATCH;
		add_current_line_type D.MINUS; token lexbuf) )
# 2943 "parsing_cocci/lexer_cocci.ml"

  | 14 ->
# 669 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true; TPlus (get_current_line_type lexbuf))
          else (patch_or_match PATCH;
		add_current_line_type D.PLUS; token lexbuf) )
# 2952 "parsing_cocci/lexer_cocci.ml"

  | 15 ->
# 674 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true; TWhy (get_current_line_type lexbuf))
          else (add_current_line_type D.OPT; token lexbuf) )
# 2960 "parsing_cocci/lexer_cocci.ml"

  | 16 ->
# 678 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TBang (get_current_line_type lexbuf) )
# 2965 "parsing_cocci/lexer_cocci.ml"

  | 17 ->
# 679 "parsing_cocci/lexer_cocci.mll"
        ( if not !col_zero
	  then (start_line true; TOPar (get_current_line_type lexbuf))
          else
            (start_line true; check_context_linetype (tok lexbuf);
	     TOPar0 ("(",get_current_line_type lexbuf)))
# 2974 "parsing_cocci/lexer_cocci.ml"

  | 18 ->
# 684 "parsing_cocci/lexer_cocci.mll"
          ( start_line true;
	    TOPar0 ("\\(",contextify(get_current_line_type lexbuf)) )
# 2980 "parsing_cocci/lexer_cocci.ml"

  | 19 ->
# 686 "parsing_cocci/lexer_cocci.mll"
        ( if not (!col_zero)
	  then (start_line true; TOr(get_current_line_type lexbuf))
          else (start_line true;
		check_context_linetype (tok lexbuf);
		TMid0 ("|",get_current_line_type lexbuf)))
# 2989 "parsing_cocci/lexer_cocci.ml"

  | 20 ->
# 691 "parsing_cocci/lexer_cocci.mll"
          ( start_line true;
	    TMid0 ("\\|",contextify(get_current_line_type lexbuf)) )
# 2995 "parsing_cocci/lexer_cocci.ml"

  | 21 ->
# 693 "parsing_cocci/lexer_cocci.mll"
        ( if not !col_zero
	  then (start_line true; TCPar (get_current_line_type lexbuf))
          else
            (start_line true; check_context_linetype (tok lexbuf);
	     TCPar0 (")",get_current_line_type lexbuf)))
# 3004 "parsing_cocci/lexer_cocci.ml"

  | 22 ->
# 698 "parsing_cocci/lexer_cocci.mll"
          ( start_line true;
	    TCPar0 ("\\)",contextify(get_current_line_type lexbuf)) )
# 3010 "parsing_cocci/lexer_cocci.ml"

  | 23 ->
# 701 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOCro (get_current_line_type lexbuf)   )
# 3015 "parsing_cocci/lexer_cocci.ml"

  | 24 ->
# 702 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCCro (get_current_line_type lexbuf)   )
# 3020 "parsing_cocci/lexer_cocci.ml"

  | 25 ->
# 703 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOBrace (get_current_line_type lexbuf) )
# 3025 "parsing_cocci/lexer_cocci.ml"

  | 26 ->
# 704 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCBrace (get_current_line_type lexbuf) )
# 3030 "parsing_cocci/lexer_cocci.ml"

  | 27 ->
# 706 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TPtrOp (get_current_line_type lexbuf)  )
# 3035 "parsing_cocci/lexer_cocci.ml"

  | 28 ->
# 707 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TDot (get_current_line_type lexbuf)    )
# 3040 "parsing_cocci/lexer_cocci.ml"

  | 29 ->
# 708 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TComma (get_current_line_type lexbuf)  )
# 3045 "parsing_cocci/lexer_cocci.ml"

  | 30 ->
# 709 "parsing_cocci/lexer_cocci.mll"
               ( start_line true; TVAEllipsis (get_current_line_type lexbuf) )
# 3050 "parsing_cocci/lexer_cocci.ml"

  | 31 ->
# 710 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TPtVirg (get_current_line_type lexbuf) )
# 3055 "parsing_cocci/lexer_cocci.ml"

  | 32 ->
# 713 "parsing_cocci/lexer_cocci.mll"
                   ( pass_zero();
		     if !current_line_started
		     then
		       (start_line true; TMul (get_current_line_type lexbuf))
		     else
		       (patch_or_match MATCH;
			add_current_line_type D.MINUS; token lexbuf) )
# 3066 "parsing_cocci/lexer_cocci.ml"

  | 33 ->
# 720 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TDmOp (Ast.Div,get_current_line_type lexbuf) )
# 3072 "parsing_cocci/lexer_cocci.ml"

  | 34 ->
# 722 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true;
		     TDmOp (Ast.Min,get_current_line_type lexbuf) )
# 3078 "parsing_cocci/lexer_cocci.ml"

  | 35 ->
# 724 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true;
		     TDmOp (Ast.Max,get_current_line_type lexbuf) )
# 3084 "parsing_cocci/lexer_cocci.ml"

  | 36 ->
# 726 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TDmOp (Ast.Mod,get_current_line_type lexbuf) )
# 3090 "parsing_cocci/lexer_cocci.ml"

  | 37 ->
# 728 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;  TTilde (get_current_line_type lexbuf) )
# 3095 "parsing_cocci/lexer_cocci.ml"

  | 38 ->
# 730 "parsing_cocci/lexer_cocci.mll"
                   ( pass_zero();
 		     if !current_line_started
 		     then
 		       (start_line true; TInc (get_current_line_type lexbuf))
 		     else (patch_or_match PATCH;
 			   add_current_line_type D.PLUSPLUS; token lexbuf) )
# 3105 "parsing_cocci/lexer_cocci.ml"

  | 39 ->
# 736 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;  TDec (get_current_line_type lexbuf) )
# 3110 "parsing_cocci/lexer_cocci.ml"

  | 40 ->
# 738 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TEq (get_current_line_type lexbuf) )
# 3115 "parsing_cocci/lexer_cocci.ml"

  | 41 ->
# 740 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Minus lexbuf )
# 3120 "parsing_cocci/lexer_cocci.ml"

  | 42 ->
# 741 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Plus lexbuf )
# 3125 "parsing_cocci/lexer_cocci.ml"

  | 43 ->
# 743 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mul lexbuf )
# 3130 "parsing_cocci/lexer_cocci.ml"

  | 44 ->
# 744 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Div lexbuf )
# 3135 "parsing_cocci/lexer_cocci.ml"

  | 45 ->
# 745 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mod lexbuf )
# 3140 "parsing_cocci/lexer_cocci.ml"

  | 46 ->
# 747 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.And lexbuf )
# 3145 "parsing_cocci/lexer_cocci.ml"

  | 47 ->
# 748 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Or lexbuf )
# 3150 "parsing_cocci/lexer_cocci.ml"

  | 48 ->
# 749 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Xor lexbuf )
# 3155 "parsing_cocci/lexer_cocci.ml"

  | 49 ->
# 750 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; mkassign Ast.Max lexbuf )
# 3160 "parsing_cocci/lexer_cocci.ml"

  | 50 ->
# 751 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; mkassign Ast.Min lexbuf )
# 3165 "parsing_cocci/lexer_cocci.ml"

  | 51 ->
# 753 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecLeft lexbuf )
# 3170 "parsing_cocci/lexer_cocci.ml"

  | 52 ->
# 754 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecRight lexbuf )
# 3175 "parsing_cocci/lexer_cocci.ml"

  | 53 ->
# 756 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TDotDot (get_current_line_type lexbuf) )
# 3180 "parsing_cocci/lexer_cocci.ml"

  | 54 ->
# 758 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TEqEq    (get_current_line_type lexbuf) )
# 3185 "parsing_cocci/lexer_cocci.ml"

  | 55 ->
# 759 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TNotEq   (get_current_line_type lexbuf) )
# 3190 "parsing_cocci/lexer_cocci.ml"

  | 56 ->
# 760 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.SupEq,get_current_line_type lexbuf) )
# 3196 "parsing_cocci/lexer_cocci.ml"

  | 57 ->
# 762 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.InfEq,get_current_line_type lexbuf) )
# 3202 "parsing_cocci/lexer_cocci.ml"

  | 58 ->
# 764 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.Inf,get_current_line_type lexbuf) )
# 3208 "parsing_cocci/lexer_cocci.ml"

  | 59 ->
# 766 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TLogOp(Ast.Sup,get_current_line_type lexbuf) )
# 3214 "parsing_cocci/lexer_cocci.ml"

  | 60 ->
# 769 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TAndLog (get_current_line_type lexbuf) )
# 3219 "parsing_cocci/lexer_cocci.ml"

  | 61 ->
# 770 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TOrLog  (get_current_line_type lexbuf) )
# 3224 "parsing_cocci/lexer_cocci.ml"

  | 62 ->
# 772 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TShROp(Ast.DecRight,get_current_line_type lexbuf) )
# 3230 "parsing_cocci/lexer_cocci.ml"

  | 63 ->
# 774 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TShLOp(Ast.DecLeft,get_current_line_type lexbuf) )
# 3236 "parsing_cocci/lexer_cocci.ml"

  | 64 ->
# 777 "parsing_cocci/lexer_cocci.mll"
                   ( if not (!col_zero)
                     then (start_line true; TAnd(get_current_line_type lexbuf))
                     else (start_line true;
	                   check_context_linetype (tok lexbuf);
		           TAnd0 ("&",get_current_line_type lexbuf)))
# 3245 "parsing_cocci/lexer_cocci.ml"

  | 65 ->
# 782 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
	             TAnd0 ("\\&",contextify(get_current_line_type lexbuf)) )
# 3251 "parsing_cocci/lexer_cocci.ml"

  | 66 ->
# 785 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TXor(get_current_line_type lexbuf) )
# 3256 "parsing_cocci/lexer_cocci.ml"

  | 67 ->
# 787 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; TCppConcatOp )
# 3261 "parsing_cocci/lexer_cocci.ml"

  | 68 ->
let
# 788 "parsing_cocci/lexer_cocci.mll"
                                               wss
# 3267 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 788 "parsing_cocci/lexer_cocci.mll"
                                                        def
# 3272 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(1)
and
# 789 "parsing_cocci/lexer_cocci.mll"
                                  ident
# 3277 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_curr_pos in
# 790 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off = String.length def in
	(* -1 in the code below because the ident is not at the line start *)
	TUndef
	  (lt,
	   check_var ident
	     (arity,line,lline,llend,offset+off,col+off,[],[],[],wss)) )
# 3289 "parsing_cocci/lexer_cocci.ml"

  | 69 ->
let
# 799 "parsing_cocci/lexer_cocci.mll"
                                                  wss
# 3295 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 799 "parsing_cocci/lexer_cocci.mll"
                                                            def
# 3300 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(1)
and
# 800 "parsing_cocci/lexer_cocci.mll"
                                   ident
# 3305 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_curr_pos in
# 801 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off = String.length def in
	(* -1 in the code below because the ident is not at the line start *)
	TDefine
	  (lt,
	   check_var ident
	     (arity,line,lline,llend,offset+off,col+off,[],[],[],wss)) )
# 3317 "parsing_cocci/lexer_cocci.ml"

  | 70 ->
let
# 810 "parsing_cocci/lexer_cocci.mll"
                                                  wss
# 3323 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 810 "parsing_cocci/lexer_cocci.mll"
                                                            def
# 3328 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(1)
and
# 811 "parsing_cocci/lexer_cocci.mll"
                                    ident
# 3333 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) (lexbuf.Lexing.lex_curr_pos + -1) in
# 813 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off = String.length def in
	TDefineParam
        (lt,
	 check_var ident
	   (* why pos here but not above? *)
	   (arity,line,lline,llend,offset+off,col+off,strbef,straft,pos,wss),
	 offset + off + (String.length ident),
	 col + off + (String.length ident)) )
# 3347 "parsing_cocci/lexer_cocci.ml"

  | 71 ->
let
# 824 "parsing_cocci/lexer_cocci.mll"
                                    prag
# 3353 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_mem.(0)
and
# 824 "parsing_cocci/lexer_cocci.mll"
                                                          wss1
# 3358 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 825 "parsing_cocci/lexer_cocci.mll"
                                    ident
# 3363 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(1) lexbuf.Lexing.lex_mem.(2)
and
# 825 "parsing_cocci/lexer_cocci.mll"
                                                           wss2
# 3368 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3)
and
# 825 "parsing_cocci/lexer_cocci.mll"
                                                                               rest
# 3373 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_curr_pos in
# 826 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let (arity,line,lline,llend,offset,col,strbef,straft,pos,ws) as lt =
	  get_current_line_type lexbuf in
	let off1 = String.length prag + String.length wss1 in
	let off2 = off1 + String.length ident + String.length wss2 in
	TPragma(lt,
		check_var ident
	        (* why pos here but not above? *)
		  (arity,line,lline,llend,offset+off1,col+off1,strbef,straft,
		   pos,wss1),
		rest,
		(arity,line,lline,llend,offset+off2,col+off2,strbef,straft,
		 pos,wss2)) )
# 3389 "parsing_cocci/lexer_cocci.ml"

  | 72 ->
# 842 "parsing_cocci/lexer_cocci.mll"
      ( TIncludeL
	  (let str = tok lexbuf in
	  let start = String.index str '\"' in
	  let finish = String.rindex str '\"' in
	  start_line true;
	  (process_include start finish str, get_current_line_type lexbuf)) )
# 3399 "parsing_cocci/lexer_cocci.ml"

  | 73 ->
# 849 "parsing_cocci/lexer_cocci.mll"
      ( TIncludeNL
	  (let str = tok lexbuf in
	  let start = String.index str '<' in
	  let finish = String.rindex str '>' in
	  start_line true;
	  (process_include start finish str,get_current_line_type lexbuf)) )
# 3409 "parsing_cocci/lexer_cocci.ml"

  | 74 ->
# 856 "parsing_cocci/lexer_cocci.mll"
      ( check_minus_context_linetype (tok lexbuf);
	start_line true;
	TIncludeAny("...",get_current_line_type lexbuf) )
# 3416 "parsing_cocci/lexer_cocci.ml"

  | 75 ->
# 860 "parsing_cocci/lexer_cocci.mll"
      ( check_context_linetype (tok lexbuf);
	start_line true;
	TInclude(get_current_line_type lexbuf) )
# 3423 "parsing_cocci/lexer_cocci.ml"

  | 76 ->
# 871 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; check_plus_linetype (tok lexbuf);
	TDirective (Ast.Noindent(tok lexbuf), get_current_line_type lexbuf) )
# 3429 "parsing_cocci/lexer_cocci.ml"

  | 77 ->
# 875 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	let res = TCppEscapedNewline (get_current_line_type lexbuf) in
	reset_line lexbuf;
	res )
# 3437 "parsing_cocci/lexer_cocci.ml"

  | 78 ->
# 881 "parsing_cocci/lexer_cocci.mll"
      (
       match !current_line_type with
        (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
        start_line true;
	(* second argument to TDirective is not quite right, because
	   it represents only the first token of the comment, but that
	   should be good enough *)
	TDirective (Ast.Indent("/*"^(comment check_comment lexbuf)),
		 get_current_line_type lexbuf)
      |	_ -> let _ = comment (fun _ -> ()) lexbuf in token lexbuf )
# 3451 "parsing_cocci/lexer_cocci.ml"

  | 79 ->
# 892 "parsing_cocci/lexer_cocci.mll"
      ( (if !current_line_started
      then lexerr "--- must be at the beginning of the line" "");
	start_line true;
	TMinusFile
	  (let str = tok lexbuf in
	  (drop_spaces(String.sub str 3 (String.length str - 3)),
	   (get_current_line_type lexbuf))) )
# 3462 "parsing_cocci/lexer_cocci.ml"

  | 80 ->
# 900 "parsing_cocci/lexer_cocci.mll"
      ( (if !current_line_started
      then lexerr "+++ must be at the beginning of the line" "");
	start_line true;
	TPlusFile
	  (let str = tok lexbuf in
	  (drop_spaces(String.sub str 3 (String.length str - 3)),
	   (get_current_line_type lexbuf))) )
# 3473 "parsing_cocci/lexer_cocci.ml"

  | 81 ->
# 909 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3478 "parsing_cocci/lexer_cocci.ml"

  | 82 ->
# 913 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3483 "parsing_cocci/lexer_cocci.ml"

  | 83 ->
# 920 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3495 "parsing_cocci/lexer_cocci.ml"

  | 84 ->
# 931 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3507 "parsing_cocci/lexer_cocci.ml"

  | 85 ->
# 947 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3519 "parsing_cocci/lexer_cocci.ml"

  | 86 ->
# 960 "parsing_cocci/lexer_cocci.mll"
      (
	start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3531 "parsing_cocci/lexer_cocci.ml"

  | 87 ->
# 971 "parsing_cocci/lexer_cocci.mll"
        ( start_line true;
	  TChar(char lexbuf,get_current_line_type lexbuf) )
# 3537 "parsing_cocci/lexer_cocci.ml"

  | 88 ->
# 973 "parsing_cocci/lexer_cocci.mll"
         ( start_line true;
	  TString(string lexbuf,(get_current_line_type lexbuf)) )
# 3543 "parsing_cocci/lexer_cocci.ml"

  | 89 ->
let
# 975 "parsing_cocci/lexer_cocci.mll"
             x
# 3549 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 975 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TFloat(x,(get_current_line_type lexbuf)) )
# 3554 "parsing_cocci/lexer_cocci.ml"

  | 90 ->
let
# 986 "parsing_cocci/lexer_cocci.mll"
         x
# 3560 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 986 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; TInt(x,(get_current_line_type lexbuf)) )
# 3564 "parsing_cocci/lexer_cocci.ml"

  | 91 ->
let
# 988 "parsing_cocci/lexer_cocci.mll"
                           x
# 3570 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 989 "parsing_cocci/lexer_cocci.mll"
      ( if !Flag.ibm
      then
	begin
	  start_line true;
	  let len = string_of_int(String.length x - 1) in
          TDecimalCst(x,len,"0",(get_current_line_type lexbuf))
	end
      else failwith "unrecognized constant modifier d/D" )
# 3581 "parsing_cocci/lexer_cocci.ml"

  | 92 ->
# 998 "parsing_cocci/lexer_cocci.mll"
                   ( TIso )
# 3586 "parsing_cocci/lexer_cocci.ml"

  | 93 ->
# 999 "parsing_cocci/lexer_cocci.mll"
                   ( TRightIso )
# 3591 "parsing_cocci/lexer_cocci.ml"

  | 94 ->
# 1001 "parsing_cocci/lexer_cocci.mll"
                   ( EOF )
# 3596 "parsing_cocci/lexer_cocci.ml"

  | 95 ->
# 1003 "parsing_cocci/lexer_cocci.mll"
      ( lexerr "unrecognised symbol, in token rule: " (tok lexbuf) )
# 3601 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and metavariable_decl_token lexbuf =
   __ocaml_lex_metavariable_decl_token_rec lexbuf 235
and __ocaml_lex_metavariable_decl_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1008 "parsing_cocci/lexer_cocci.mll"
    ( reset_line lexbuf; metavariable_decl_token lexbuf )
# 3613 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
# 1010 "parsing_cocci/lexer_cocci.mll"
                  (
    start_line false; metavariable_decl_token lexbuf )
# 3619 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
# 1013 "parsing_cocci/lexer_cocci.mll"
                                   (
    start_line false; metavariable_decl_token lexbuf )
# 3625 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
# 1016 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TArobArob )
# 3630 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
# 1018 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TTildeEq (get_current_line_type lexbuf) )
# 3635 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 1019 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TTildeExclEq (get_current_line_type lexbuf) )
# 3640 "parsing_cocci/lexer_cocci.ml"

  | 6 ->
# 1020 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TEq (get_current_line_type lexbuf) )
# 3645 "parsing_cocci/lexer_cocci.ml"

  | 7 ->
# 1021 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOPar (get_current_line_type lexbuf) )
# 3650 "parsing_cocci/lexer_cocci.ml"

  | 8 ->
# 1022 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCPar (get_current_line_type lexbuf) )
# 3655 "parsing_cocci/lexer_cocci.ml"

  | 9 ->
# 1024 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOCro (get_current_line_type lexbuf)   )
# 3660 "parsing_cocci/lexer_cocci.ml"

  | 10 ->
# 1025 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCCro (get_current_line_type lexbuf)   )
# 3665 "parsing_cocci/lexer_cocci.ml"

  | 11 ->
# 1026 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TOBrace (get_current_line_type lexbuf) )
# 3670 "parsing_cocci/lexer_cocci.ml"

  | 12 ->
# 1027 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TCBrace (get_current_line_type lexbuf) )
# 3675 "parsing_cocci/lexer_cocci.ml"

  | 13 ->
# 1029 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TPtrOp (get_current_line_type lexbuf)  )
# 3680 "parsing_cocci/lexer_cocci.ml"

  | 14 ->
# 1030 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TDot (get_current_line_type lexbuf)    )
# 3685 "parsing_cocci/lexer_cocci.ml"

  | 15 ->
# 1031 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TComma (get_current_line_type lexbuf)  )
# 3690 "parsing_cocci/lexer_cocci.ml"

  | 16 ->
# 1032 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TMPtVirg (* works better with tokens_all *) )
# 3696 "parsing_cocci/lexer_cocci.ml"

  | 17 ->
# 1034 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true;
		     TShLOp(Ast.DecLeft,get_current_line_type lexbuf) )
# 3702 "parsing_cocci/lexer_cocci.ml"

  | 18 ->
# 1037 "parsing_cocci/lexer_cocci.mll"
                   ( pass_zero();
		     if !current_line_started
		     then
		       (start_line true; TMul (get_current_line_type lexbuf))
		     else
		       (patch_or_match MATCH;
			add_current_line_type D.MINUS;
			metavariable_decl_token lexbuf) )
# 3714 "parsing_cocci/lexer_cocci.ml"

  | 19 ->
# 1046 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TEqEq    (get_current_line_type lexbuf) )
# 3719 "parsing_cocci/lexer_cocci.ml"

  | 20 ->
# 1047 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TNotEq   (get_current_line_type lexbuf) )
# 3724 "parsing_cocci/lexer_cocci.ml"

  | 21 ->
# 1048 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; TSub     (get_current_line_type lexbuf) )
# 3729 "parsing_cocci/lexer_cocci.ml"

  | 22 ->
# 1049 "parsing_cocci/lexer_cocci.mll"
        ( (start_line true; TPlus (get_current_line_type lexbuf)) )
# 3734 "parsing_cocci/lexer_cocci.ml"

  | 23 ->
# 1050 "parsing_cocci/lexer_cocci.mll"
        ( (start_line true; TMinus (get_current_line_type lexbuf)) )
# 3739 "parsing_cocci/lexer_cocci.ml"

  | 24 ->
# 1051 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TDmOp (Ast.Div,get_current_line_type lexbuf) )
# 3744 "parsing_cocci/lexer_cocci.ml"

  | 25 ->
# 1052 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TDmOp (Ast.Mod,get_current_line_type lexbuf) )
# 3749 "parsing_cocci/lexer_cocci.ml"

  | 26 ->
# 1053 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TShROp(Ast.DecRight,get_current_line_type lexbuf) )
# 3754 "parsing_cocci/lexer_cocci.ml"

  | 27 ->
# 1054 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TAnd (get_current_line_type lexbuf) )
# 3759 "parsing_cocci/lexer_cocci.ml"

  | 28 ->
# 1055 "parsing_cocci/lexer_cocci.mll"
        (  (start_line true; TOr(get_current_line_type lexbuf)) )
# 3764 "parsing_cocci/lexer_cocci.ml"

  | 29 ->
# 1056 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TXor(get_current_line_type lexbuf) )
# 3769 "parsing_cocci/lexer_cocci.ml"

  | 30 ->
# 1057 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TLogOp(Ast.SupEq,get_current_line_type lexbuf) )
# 3774 "parsing_cocci/lexer_cocci.ml"

  | 31 ->
# 1058 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TLogOp(Ast.Inf,get_current_line_type lexbuf) )
# 3779 "parsing_cocci/lexer_cocci.ml"

  | 32 ->
# 1059 "parsing_cocci/lexer_cocci.mll"
        ( start_line true; TLogOp(Ast.Sup,get_current_line_type lexbuf) )
# 3784 "parsing_cocci/lexer_cocci.ml"

  | 33 ->
# 1060 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TAndLog (get_current_line_type lexbuf) )
# 3789 "parsing_cocci/lexer_cocci.ml"

  | 34 ->
# 1061 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TOrLog (get_current_line_type lexbuf) )
# 3794 "parsing_cocci/lexer_cocci.ml"

  | 35 ->
# 1062 "parsing_cocci/lexer_cocci.mll"
         ( start_line true; TDotDot (get_current_line_type lexbuf) )
# 3799 "parsing_cocci/lexer_cocci.ml"

  | 36 ->
# 1063 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Minus lexbuf )
# 3804 "parsing_cocci/lexer_cocci.ml"

  | 37 ->
# 1064 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Plus lexbuf )
# 3809 "parsing_cocci/lexer_cocci.ml"

  | 38 ->
# 1065 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mul lexbuf )
# 3814 "parsing_cocci/lexer_cocci.ml"

  | 39 ->
# 1066 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Div lexbuf )
# 3819 "parsing_cocci/lexer_cocci.ml"

  | 40 ->
# 1067 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Mod lexbuf )
# 3824 "parsing_cocci/lexer_cocci.ml"

  | 41 ->
# 1068 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.And lexbuf )
# 3829 "parsing_cocci/lexer_cocci.ml"

  | 42 ->
# 1069 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Or lexbuf )
# 3834 "parsing_cocci/lexer_cocci.ml"

  | 43 ->
# 1070 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.Xor lexbuf )
# 3839 "parsing_cocci/lexer_cocci.ml"

  | 44 ->
# 1071 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecLeft lexbuf )
# 3844 "parsing_cocci/lexer_cocci.ml"

  | 45 ->
# 1072 "parsing_cocci/lexer_cocci.mll"
                   ( start_line true; mkassign Ast.DecRight lexbuf )
# 3849 "parsing_cocci/lexer_cocci.ml"

  | 46 ->
# 1074 "parsing_cocci/lexer_cocci.mll"
      (match !current_line_type with
        (D.PLUS,_,_) | (D.PLUSPLUS,_,_) ->
        start_line true;
	(* second argument to TDirective is not quite right, because
	   it represents only the first token of the comment, but that
	   should be good enough *)
	TDirective (Ast.Indent("/*"^(comment check_comment lexbuf)),
		 get_current_line_type lexbuf)
      |	_ -> let _ = comment (fun _ -> ()) lexbuf in
	     metavariable_decl_token lexbuf )
# 3863 "parsing_cocci/lexer_cocci.ml"

  | 47 ->
# 1085 "parsing_cocci/lexer_cocci.mll"
                    ( start_line true; TCppConcatOp (* for fresh vars *) )
# 3868 "parsing_cocci/lexer_cocci.ml"

  | 48 ->
# 1088 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3873 "parsing_cocci/lexer_cocci.ml"

  | 49 ->
# 1092 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; id_tokens lexbuf )
# 3878 "parsing_cocci/lexer_cocci.ml"

  | 50 ->
# 1099 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3889 "parsing_cocci/lexer_cocci.ml"

  | 51 ->
# 1109 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "< and > not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3900 "parsing_cocci/lexer_cocci.ml"

  | 52 ->
# 1124 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3911 "parsing_cocci/lexer_cocci.ml"

  | 53 ->
# 1136 "parsing_cocci/lexer_cocci.mll"
      ( start_line true;
	if not !Flag.c_plus_plus
	then
	  Common.pr2_once
	    "~ and :: not allowed in C identifiers, try -c++ option";
	id_tokens lexbuf
      )
# 3922 "parsing_cocci/lexer_cocci.ml"

  | 54 ->
# 1146 "parsing_cocci/lexer_cocci.mll"
        ( start_line true;
	  TChar(char lexbuf,get_current_line_type lexbuf) )
# 3928 "parsing_cocci/lexer_cocci.ml"

  | 55 ->
# 1148 "parsing_cocci/lexer_cocci.mll"
         ( start_line true;
	  TString(string lexbuf,(get_current_line_type lexbuf)) )
# 3934 "parsing_cocci/lexer_cocci.ml"

  | 56 ->
let
# 1150 "parsing_cocci/lexer_cocci.mll"
             x
# 3940 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1150 "parsing_cocci/lexer_cocci.mll"
                   ( TFloat(x,(get_current_line_type lexbuf)) )
# 3944 "parsing_cocci/lexer_cocci.ml"

  | 57 ->
let
# 1159 "parsing_cocci/lexer_cocci.mll"
         x
# 3950 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1159 "parsing_cocci/lexer_cocci.mll"
            ( start_line true; TInt(x,(get_current_line_type lexbuf)) )
# 3954 "parsing_cocci/lexer_cocci.ml"

  | 58 ->
let
# 1161 "parsing_cocci/lexer_cocci.mll"
                           x
# 3960 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1162 "parsing_cocci/lexer_cocci.mll"
      ( if !Flag.ibm
      then
	begin
	  start_line true;
	  let len = string_of_int(String.length x - 1) in
          TDecimalCst(x,len,"0",(get_current_line_type lexbuf))
	end
      else failwith "unrecognized constant modifier d/D" )
# 3971 "parsing_cocci/lexer_cocci.ml"

  | 59 ->
# 1171 "parsing_cocci/lexer_cocci.mll"
      ( lexerr
	 "metavariables: unrecognised symbol in metavariable_decl_token rule: "
	 (tok lexbuf) )
# 3978 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_metavariable_decl_token_rec lexbuf __ocaml_lex_state

and char lexbuf =
   __ocaml_lex_char_rec lexbuf 343
and __ocaml_lex_char_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 1177 "parsing_cocci/lexer_cocci.mll"
          x
# 3991 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1177 "parsing_cocci/lexer_cocci.mll"
                                       ( String.make 1 x ^ restchars lexbuf )
# 3995 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 1179 "parsing_cocci/lexer_cocci.mll"
                                             x
# 4001 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1179 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4005 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 1183 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4011 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1183 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4015 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 1184 "parsing_cocci/lexer_cocci.mll"
                 v
# 4021 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 1184 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4026 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 1185 "parsing_cocci/lexer_cocci.mll"
 (
          (match v with (* Machine specific ? *)
          | 'n' -> ()  | 't' -> ()   | 'v' -> ()  | 'b' -> () | 'r' -> ()
          | 'f' -> () | 'a' -> ()
	  | '\\' -> () | '?'  -> () | '\'' -> ()  | '\"' -> ()
          | 'e' -> () (* linuxext: ? *)
	  | _ ->
              Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
	  );
          x ^ restchars lexbuf
	)
# 4040 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
# 1197 "parsing_cocci/lexer_cocci.mll"
      ( Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
        tok lexbuf ^ restchars lexbuf
      )
# 4047 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_char_rec lexbuf __ocaml_lex_state

and restchars lexbuf =
   __ocaml_lex_restchars_rec lexbuf 353
and __ocaml_lex_restchars_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1202 "parsing_cocci/lexer_cocci.mll"
                                       ( "" )
# 4059 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 1203 "parsing_cocci/lexer_cocci.mll"
          x
# 4065 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1203 "parsing_cocci/lexer_cocci.mll"
                                       ( String.make 1 x ^ restchars lexbuf )
# 4069 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 1205 "parsing_cocci/lexer_cocci.mll"
                                             x
# 4075 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1205 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4079 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 1209 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4085 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1209 "parsing_cocci/lexer_cocci.mll"
                                                     ( x ^ restchars lexbuf )
# 4089 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
let
# 1210 "parsing_cocci/lexer_cocci.mll"
                 v
# 4095 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 1210 "parsing_cocci/lexer_cocci.mll"
                                  x
# 4100 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 1211 "parsing_cocci/lexer_cocci.mll"
 (
          (match v with (* Machine specific ? *)
          | 'n' -> ()  | 't' -> ()   | 'v' -> ()  | 'b' -> () | 'r' -> ()
          | 'f' -> () | 'a' -> ()
	  | '\\' -> () | '?'  -> () | '\'' -> ()  | '\"' -> ()
          | 'e' -> () (* linuxext: ? *)
	  | _ ->
              Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
	  );
          x ^ restchars lexbuf
	)
# 4114 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 1223 "parsing_cocci/lexer_cocci.mll"
      ( Common.pr2 ("LEXER: unrecognised symbol in char:"^tok lexbuf);
        tok lexbuf ^ restchars lexbuf
      )
# 4121 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_restchars_rec lexbuf __ocaml_lex_state

and string lexbuf =
   __ocaml_lex_string_rec lexbuf 364
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1228 "parsing_cocci/lexer_cocci.mll"
                                               ( "" )
# 4133 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
let
# 1229 "parsing_cocci/lexer_cocci.mll"
                                 x
# 4139 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1230 "parsing_cocci/lexer_cocci.mll"
    ( line := !line + 1; (Printf.sprintf "%c" x) ^ string lexbuf )
# 4143 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
let
# 1231 "parsing_cocci/lexer_cocci.mll"
          x
# 4149 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 1231 "parsing_cocci/lexer_cocci.mll"
                               ( Common.string_of_char x ^ string lexbuf )
# 4153 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
let
# 1232 "parsing_cocci/lexer_cocci.mll"
                                            x
# 4159 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1232 "parsing_cocci/lexer_cocci.mll"
                                              ( x ^ string lexbuf )
# 4163 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
let
# 1233 "parsing_cocci/lexer_cocci.mll"
                               x
# 4169 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos lexbuf.Lexing.lex_curr_pos in
# 1233 "parsing_cocci/lexer_cocci.mll"
                                              ( x ^ string lexbuf )
# 4173 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
let
# 1234 "parsing_cocci/lexer_cocci.mll"
                v
# 4179 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 1)
and
# 1234 "parsing_cocci/lexer_cocci.mll"
                       x
# 4184 "parsing_cocci/lexer_cocci.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 1235 "parsing_cocci/lexer_cocci.mll"
       (
         (match v with
	    | 'n' -> ()  | 't' -> ()   | 'v' -> ()  | 'b' -> () | 'r' -> ()
	    | 'f' -> () | 'a' -> ()
	    | '\\' -> () | '?'  -> () | '\'' -> ()  | '\"' -> ()
	    | 'e' -> ()
	    | '\n' -> ()
	    | '(' -> () | '|' -> () | ')' -> ()
	    | _ -> lexerr "unrecognised symbol:" (tok lexbuf)
	 );
          x ^ string lexbuf
       )
# 4199 "parsing_cocci/lexer_cocci.ml"

  | 6 ->
# 1247 "parsing_cocci/lexer_cocci.mll"
      ( lexerr "unrecognised symbol: " (tok lexbuf) )
# 4204 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and comment check_comment lexbuf =
   __ocaml_lex_comment_rec check_comment lexbuf 376
and __ocaml_lex_comment_rec check_comment lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 1250 "parsing_cocci/lexer_cocci.mll"
         ( let s = tok lexbuf in check_comment s; start_line true; s )
# 4216 "parsing_cocci/lexer_cocci.ml"

  | 1 ->
# 1252 "parsing_cocci/lexer_cocci.mll"
      ( let s = tok lexbuf in
        (* even blank line should have a + *)
        check_comment s;
        reset_line lexbuf; s ^ comment check_comment lexbuf )
# 4224 "parsing_cocci/lexer_cocci.ml"

  | 2 ->
# 1256 "parsing_cocci/lexer_cocci.mll"
        ( pass_zero();
	  if !current_line_started
	  then (start_line true;
		let s = tok lexbuf in s^(comment check_comment lexbuf))
	  else (start_line true; comment check_comment lexbuf) )
# 4233 "parsing_cocci/lexer_cocci.ml"

  | 3 ->
# 1263 "parsing_cocci/lexer_cocci.mll"
      ( let s = tok lexbuf in
        check_comment s; start_line true; s ^ comment check_comment lexbuf )
# 4239 "parsing_cocci/lexer_cocci.ml"

  | 4 ->
# 1266 "parsing_cocci/lexer_cocci.mll"
      ( let s = tok lexbuf in
        check_comment s; start_line true; s ^ comment check_comment lexbuf )
# 4245 "parsing_cocci/lexer_cocci.ml"

  | 5 ->
# 1269 "parsing_cocci/lexer_cocci.mll"
      ( start_line true; let s = tok lexbuf in
        Common.pr2 ("LEXER: unrecognised symbol in comment:"^s);
        s ^ comment check_comment lexbuf
      )
# 4253 "parsing_cocci/lexer_cocci.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf;
      __ocaml_lex_comment_rec check_comment lexbuf __ocaml_lex_state

;;

